#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      ic.py
#  brief:     generating initial conditions
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################


from pNbody import *
from pNbody import iclib
import numpy as np

import sys

try:
    from scipy.integrate import quadrature
    from scipy import special
    from scipy import optimize
    is_scipy = True
except ImportError:
    is_scipy = False


"""
# isotropic velocities
random2 = np.random.random([n])
random3 = np.random.random([n])
p = 2*pi*random2
costh = 1.-2.*random3
sinth = np.sqrt(1.-costh**2)
vx = v*sinth*np.cos(p)
vy = v*sinth*np.sin(p)
vz = v*costh

# radial velocities
x = nb.pos[:,0]
y = nb.pos[:,1]
z = nb.pos[:,2]
vx = v*x/r
vy = v*y/r
vz = v*z/r



Some notes on the generation of initial conditions:

Spherical case
--------------

1) if M(r) is known analytically :

  1.1) if it is invertible analytically
       --> for a given x between 0 and 1, find r from M(r)

  1.2) if it is not invertible analytically


       1.2.1) use a Monte-Carlo approach
              (warning : may take time if the model is cuspy)

       1.2.2) invert M(r) numerically and create a vector of r and Mr
              then, use generic_Mr

       1.2.3) for a given x between 0 and 1, find r from M(r),
              by inverting M(r). May take time.







"""


def get_local_n(n):
    """
    This function set the global number of particle each
    node must hand.
    """

    # set the number of particles per procs
    n0 = n
    npercpu = n / mpi.mpi_NTask()

    if mpi.mpi_IsMaster():
        npercpu = npercpu + (n - npercpu * mpi.mpi_NTask())

    n = npercpu

    ntot = mpi.mpi_allreduce(n)

    if ntot != n0:
        print(("ntot=%d while n0=%d" % (ntot, n0)))
        sys.exit()

    return int(n), int(ntot)


def ComputeGridParameters(n, args, rmax, M, pr_fct, mr_fct, Neps_des, rc, ng):
    """

    This function computes dR, the appropriate grid used to approximate Mr.

    The grid is set in order to have "Neps_des" particles
    in the first division of the grid. Then, the radius of the grid
    follows an exponnential distribution up to rmax.

    1) from the density distribution, the total mass and the number of particles,
       using a newton algorithm, it computes eps, the radius that will contains "Neps_des" particles


    2) once eps is set, we determine rc (the grid scale length) from eps and ng, in order to
       have a grid with the a first cell equal to eps.


       if the computation of rc fails, we use the default value of rc


    The function takes the following arguments

    n 		: number of particles
    M		: total mass
    rmax          : max radius
    args		: list of args for the profile
    pr_fct	: profile function
    mr_fct	: mass-radius function

    Neps_des	: desired number of point in the first beam
    rc		: default size of the first beam
    ng		: number of grid divisions


    it returns :


    Rs		: grid points
    eps		: radius containing about  Neps_des particles
    Neps		: number of particles in eps
    """

    if not is_scipy:
        raise Exception(
            "module scipy needed for function ComputeModelParameters !")

    ###########################
    # some considerations
    ###########################

    # central density
    rho0 = M / mr_fct(*(rmax,) + args)

    # mass of particles
    m = M / float(n)

    args = args + (rho0,)
    rs = args[0]

    ##########################################################################
    # find eps in order to have Neps_des particles in eps
    def RfromN(r, args, m, N):
        return mr_fct(*(r,) + args) / m - N

    try:
        eps = optimize.newton(
            RfromN,
            x0=rs,
            args=(
                args,
                m,
                Neps_des),
            fprime=None,
            tol=1e-10,
            maxiter=500)
    except BaseException:
        print("fail to get eps from newton, trying bisection.")
        try:
            eps = optimize.bisection(
                RfromN,
                a=1e-10,
                b=rs,
                args=(
                    args,
                    m,
                    Neps_des),
                xtol=1e-5,
                maxiter=500)
            print("ok")
        except BaseException:
            print("fail to get eps from bisection.")
            print("quit")
            sys.exit()

    ##########################################################################

    ##########################################################################
    # compute the number of particles that will fall in eps
    Meps = mr_fct(*(eps,) + args)
    Neps = Meps / m
    ##########################################################################

    ##########################################################################
    # parameters for the adaptative grid

    # find eps in order to have Neps_des particles in eps

    def GetRc(rc, n, rmax, eps):
        return (np.exp((1. / (ng - 1)) / rc) - 1) / \
            (np.exp(1. / rc) - 1) * rmax - eps

    try:
        #rc = optimize.newton(GetRc, x0=0.1, args = (n,rmax,eps), fprime = None, tol = 1e-20, maxiter = 500)
        rc = optimize.bisection(
            GetRc,
            a=1e-4,
            b=rmax,
            args=(
                n,
                rmax,
                eps),
            xtol=1e-3,
            maxiter=500)
    except BaseException:
        print(("fail to get rc, using rc=%g." % rc))

    def gm(i): return (np.exp((i / float(ng - 1)) / rc) - 1) / \
        (np.exp(1. / rc) - 1) * rmax

    def g(r): return float(ng - 1) * rc * \
        np.log(r / rmax * (np.exp(1. / rc) - 1.) + 1.)
    Rs = gm(np.arange(ng))

    return Rs, rc, eps, Neps, g, gm


def ComputeGridParameters2(
        eps,
        nmax,
        args,
        rmax,
        M,
        pr_fct,
        mr_fct,
        Neps_des,
        rc,
        ng):
    """

    This function computes dR, the appropriate grid used to approximate Mr.

    The number of particle of the model is set in order to have "Neps_des" particles
    in the first division of the grid. Then, the radius of the grid
    follows an exponential distribution up to rmax.

    1) n is set from the total mass and Neps_des

    2) once n is set, we determine rc (the grid scale length) from eps and ng, in order to
       have a grid with the a first cell equal to eps.


       if the computation of rc fails, we use the default value of rc


    The function takes the following arguments

    eps 		: the desired grid resolution
    nmax		: max number of particles
    M		: total mass
    rmax          : max radius
    args		: list of args for the profile
    pr_fct	: profile function
    mr_fct	: mass-radius function

    Neps_des	: desired number of point in the first beam
    rc		: default size of the first beam
    ng		: number of grid divisions


    it returns :

    n		: number of particles
    Rs		: grid points
    rc		: parameter of the scaling fct
    g		: scaling fct
    gm		: inverse of scaling fct
    """

    if not is_scipy:
        raise Exception(
            "module scipy needed for function ComputeModelParameters !")

    ###########################
    # some considerations
    ###########################

    # central density
    rho0 = M / mr_fct(*(rmax,) + args)

    args = args + (rho0,)
    rs = args[0]

    # number of particles
    n = int(Neps_des * M / mr_fct(*(eps,) + args))

    # if n> nmax, find eps containing Neps_des particles
    if n > nmax:

        n = nmax

        # mass of particles
        m = M / float(n)

        #######################################################################
        # find eps in order to have Neps_des particles in eps

        def RfromN(r, args, m, N):
            return mr_fct(*(r,) + args) / m - N

        try:
            eps = optimize.newton(
                RfromN,
                x0=rs,
                args=(
                    args,
                    m,
                    Neps_des),
                fprime=None,
                tol=1e-10,
                maxiter=500)
        except BaseException:
            print("fail to get eps from newton, trying bisection.")
            try:
                eps = optimize.bisection(
                    RfromN,
                    a=1e-10,
                    b=rs,
                    args=(
                        args,
                        m,
                        Neps_des),
                    xtol=1e-5,
                    maxiter=500)
                print("ok")
            except BaseException:
                print("fail to get eps from bisection.")
                print("quit")
                sys.exit()

        #######################################################################

        #######################################################################
        # compute the number of particles that will fall in eps
        Meps = mr_fct(*(eps,) + args)
        Neps = Meps / m
        #######################################################################

    # mass of particles
    m = M / float(n)

    ##########################################################################
    # parameters for the adaptative grid

    # find eps in order to have Neps_des particles in eps

    def GetRc(rc, ng, rmax, eps):
        return (np.exp((1. / (ng - 1)) / rc) - 1) / \
            (np.exp(1. / rc) - 1) * rmax - eps

    try:
        #rc = optimize.newton(GetRc, x0=0.1, args = (n,rmax,eps), fprime = None, tol = 1e-20, maxiter = 500)
        rc = optimize.bisection(
            GetRc,
            a=1e-4,
            b=rmax,
            args=(
                ng,
                rmax,
                eps),
            xtol=1e-3,
            maxiter=500)
    except BaseException:
        print(("fail to get rc, using rc=%g." % rc))

    def gm(i): return (np.exp((i / float(ng - 1)) / rc) - 1) / \
        (np.exp(1. / rc) - 1) * rmax

    def g(r): return float(ng - 1) * rc * \
        np.log(r / rmax * (np.exp(1. / rc) - 1.) + 1.)
    Rs = gm(np.arange(ng))

    return n, eps, Rs, rc, g, gm


def invert(x, rmin, rmax, fct, args, eps=1e-10):
    """
    return vector r that corresponds to
    fct(r,args)=x
    This routine uses a simple bissector algorithm
    """
    n = len(x)
    rrmin = rmin * np.ones(n)
    rrmax = rmax * np.ones(n)
    xxmin = fct(rrmin, args) - x
    xxmax = fct(rrmax, args) - x

    if np.sum((xxmin * xxmax >= 0)) != 0:
        print("No max between rmin and rmax ! for some points")
        sys.exit()

    k = 0

    while max(abs(rrmax - rrmin)) > eps:

        print(("it = %3d err = %8.1e" % (k, max(abs(rrmax - rrmin)))))

        k = k + 1
        rr = (rrmax + rrmin) / 2.

        xx = fct(rr, args) - x

        rrmax = np.where(xxmin * xx <= 0, rr, rrmax)
        rrmin = np.where(xxmin * xx > 0, rr, rrmin)
        xxmin = np.where(xxmin * xx > 0, xx, xxmin)

    return rr


def box(n, a, b, c, irand=1, name='box.h5py', ftype='swift'):
    """
    Return an Nbody object that contains n particles distributed
    in an homogeneous box of dimension a,b,c, centred at the origin
    radius rmax.
    """

    if isinstance(n, np.ndarray):
        rand_vec = n
        n = len(n)
        ntot = mpi.mpi_allreduce(n)
    else:
        # set random seed
        np.random.seed(irand + mpi.mpi_Rank())
        # set the number of particles per procs
        n, ntot = get_local_n(n)
        # generate random numbers
        rand_vec = np.random.random([n, 3])

    pos = rand_vec - [0.5, 0.5, 0.5]
    pos = pos * np.array([2 * a, 2 * b, 2 * c])

    vel = np.ones([n, 3]) * 0.0
    mass = np.ones([n]) * 1. / ntot

    nb = Nbody(
        status='new',
        p_name=name,
        pos=pos,
        vel=vel,
        mass=mass,
        ftype=ftype)

    return nb


def homodisk(
    n,
    a,
    b,
    dz,
    irand=1,
    name='homodisk.hdf5',
    ftype='swift'):
    """
    Return an Nbody object that contains n particles distributed
    in an homogeneous oval of radius a and b, and of thickness dz.
    """

    if isinstance(n, np.ndarray):
        random1 = n[:, 0]
        random2 = n[:, 1]
        random3 = n[:, 2]
        n = len(n)
        ntot = mpi.mpi_allreduce(n)
    else:

        # set random seed
        np.random.seed(irand + mpi.mpi_Rank())
        # set the number of particles per procs
        n, ntot = get_local_n(n)

        random1 = np.random.random([n])
        random2 = np.random.random([n])
        random3 = np.random.random([n])

    xx = random1**(1. / 2.)

    theta = 2. * random2 * np.pi

    x = a * xx * np.cos(theta)
    y = b * xx * np.sin(theta)
    z = dz * random3 - dz / 2.

    pos = np.transpose(np.array([x, y, z]))
    vel = np.ones([n, 3]) * 0.0
    mass = np.ones([n]) * 1. / ntot

    nb = Nbody(
        status='new',
        p_name=name,
        pos=pos,
        vel=vel,
        mass=mass,
        ftype=ftype)

    return nb


def disklrrc(
    n,
    a,
    b,
    dz,
    irand=1,
    name='homodisk.hdf5',
    ftype='swift'):
    """
    Return an Nbody object that contains n particles distributed
    in a disk having a linear rising rotation curve.
    """

    if isinstance(n, np.ndarray):
        random1 = n[:, 0]
        random2 = n[:, 1]
        random3 = n[:, 2]
        n = len(n)
        ntot = mpi.mpi_allreduce(n)
    else:

        # set random seed
        np.random.seed(irand + mpi.mpi_Rank())
        # set the number of particles per procs
        n, ntot = get_local_n(n)

        random1 = np.random.random([n])
        random2 = np.random.random([n])
        random3 = np.random.random([n])

    xx = random1**(1. / 3)

    theta = 2. * random2 * np.pi

    x = a * xx * np.cos(theta)
    y = b * xx * np.sin(theta)
    z = dz * random3 - dz / 2.

    pos = np.transpose(np.array([x, y, z]))
    vel = np.ones([n, 3]) * 0.0
    mass = np.ones([n]) * 1. / ntot

    nb = Nbody(
        status='new',
        p_name=name,
        pos=pos,
        vel=vel,
        mass=mass,
        ftype=ftype)

    return nb


def mestel(
    n,
    rmax=1,
    dz=0,
    V0=1,
    G=1,
    irand=1,
    name='mestel.dat',
    ftype='gadget'):
    """
    Return an Nbody object that contains n particles distributed
    such as they reproduce a Mestel disk (1/R surface density).
    
    V0 : rotation curve at infinity
    """

    if isinstance(n, np.ndarray):
        random1 = n[:, 0]
        random2 = n[:, 1]
        random3 = n[:, 2]
        n = len(n)
        ntot = mpi.mpi_allreduce(n)
    else:

        # set random seed
        np.random.seed(irand + mpi.mpi_Rank())
        # set the number of particles per procs
        n, ntot = get_local_n(n)

        random1 = np.random.random([n])
        random2 = np.random.random([n])
        random3 = np.random.random([n])
    
    Mtot = V0*V0/G * rmax

    xx = random1

    theta = 2. * random2 * np.pi

    x = rmax * xx * np.cos(theta)
    y = rmax * xx * np.sin(theta)
    z = dz * random3 - dz / 2.

    pos = np.transpose(np.array([x, y, z]))
    vel = np.ones([n, 3]) * 0.0
    mass = np.ones([n]) * 1. / ntot  * Mtot

    nb = Nbody(
        status='new',
        p_name=name,
        pos=pos,
        vel=vel,
        mass=mass,
        ftype=ftype)

    return nb



def pseudomestel(
    n,
    rmax=1,
    dz=0,
    V0=1,
    a=0.1,
    G=1,
    irand=1,
    name='pseudomestel.dat',
    ftype='gadget'):
    """
    Return an Nbody object that contains n particles distributed
    such as they reproduce a Mestel disk (1/sqrt(a^2 + R^2) surface density).
    
    V0 : rotation curve at infinity
    """

    if isinstance(n, np.ndarray):
        random1 = n[:, 0]
        random2 = n[:, 1]
        random3 = n[:, 2]
        n = len(n)
        ntot = mpi.mpi_allreduce(n)
    else:

        # set random seed
        np.random.seed(irand + mpi.mpi_Rank())
        # set the number of particles per procs
        n, ntot = get_local_n(n)

        random1 = np.random.random([n])
        random2 = np.random.random([n])
        random3 = np.random.random([n])
    
    Mtot = V0*V0/G * ( np.sqrt(rmax**2 + a**2) -a   )
    xx =  np.sqrt( (G*random1*Mtot/(V0**2) + a )**2 - a**2 )
    
    theta = 2. * random2 * np.pi

    x = xx * np.cos(theta)
    y = xx * np.sin(theta)
    z = dz * random3 - dz / 2.

    pos = np.transpose(np.array([x, y, z]))
    vel = np.ones([n, 3]) * 0.0
    mass = np.ones([n]) * 1. / ntot  * Mtot

    nb = Nbody(
        status='new',
        p_name=name,
        pos=pos,
        vel=vel,
        mass=mass,
        ftype=ftype)

    return nb



def homosphere(
    n,
    a,
    b,
    c,
    irand=1,
    name='homosphere.hdf5',
    ftype='swift'):
    """
    Return an Nbody object that contains n particles distributed
    in an homogeneous triaxial sphere of axis a,b,c.
    """

    if isinstance(n, np.ndarray):
        random1 = n[:, 0]
        random2 = n[:, 1]
        random3 = n[:, 2]
        n = len(n)
        ntot = mpi.mpi_allreduce(n)
    else:

        # set random seed
        np.random.seed(irand + mpi.mpi_Rank())
        # set the number of particles per procs
        n, ntot = get_local_n(n)

        random1 = np.random.random([n])
        random2 = np.random.random([n])
        random3 = np.random.random([n])

    xm = (random1)**(1. / 3.)
    phi = random2 * np.pi * 2.
    costh = 1. - 2. * random3

    sinth = np.sqrt(1. - costh**2)
    axm = a * xm * sinth
    bxm = b * xm * sinth
    x = axm * np.cos(phi)
    y = bxm * np.sin(phi)
    z = c * xm * costh

    pos = np.transpose(np.array([x, y, z]))
    vel = np.ones([n, 3]) * 0.0
    mass = np.ones([n]) * 1. / ntot

    nb = Nbody(
        status='new',
        p_name=name,
        pos=pos,
        vel=vel,
        mass=mass,
        ftype=ftype)

    return nb


def ring(n, r, M=1.0, irand=1, name='ring.hdf5', ftype='swift'):
    """
    Ring of radius r
    """

    if isinstance(n, np.ndarray):
        random2 = n[:, 0]
        n = len(n)
        ntot = mpi.mpi_allreduce(n)
    else:
        # set random seed
        np.random.seed(irand + mpi.mpi_Rank())
        # set the number of particles per procs
        n, ntot = get_local_n(n)
        random2 = np.random.random(n)

    phi = random2 * np.pi * 2.

    x = r *np.cos(phi)
    y = r *np.sin(phi)
    z = np.zeros(n)

    pos = np.transpose(np.array([x, y, z]))
    vel = np.ones([n, 3]) * 0.0
    mass = M*np.ones([n]) * 1. / ntot

    nb = Nbody(
        status='new',
        p_name=name,
        pos=pos,
        vel=vel,
        mass=mass,
        ftype=ftype)

    return nb





def shell(n, r, irand=1, name='shell.hdf5', ftype='swift'):
    """
    Shell of radius r
    """

    if isinstance(n, np.ndarray):
        random2 = n[:, 0]
        random3 = n[:, 1]
        n = len(n)
        ntot = mpi.mpi_allreduce(n)
    else:
        # set random seed
        np.random.seed(irand + mpi.mpi_Rank())
        # set the number of particles per procs
        n, ntot = get_local_n(n)
        random2 = np.random.random(n)
        random3 = np.random.random(n)

    phi = random2 * np.pi * 2.
    costh = 1. - 2. * random3

    sinth = np.sqrt(1. - costh**2)

    x = r * sinth * np.cos(phi)
    y = r * sinth * np.sin(phi)
    z = r * costh

    pos = np.transpose(np.array([x, y, z]))
    vel = np.ones([n, 3]) * 0.0
    mass = np.ones([n]) * 1. / ntot

    nb = Nbody(
        status='new',
        p_name=name,
        pos=pos,
        vel=vel,
        mass=mass,
        ftype=ftype)

    return nb


def plummer(
    n,
    a,
    b,
    c,
    eps,
    rmax,
    M=1.,
    irand=1,
    vel='no',
    name='plummer.hdf5',
    ftype='swift'):
    """
    Return an Nbody object that contains n particles distributed
    in a triaxial plummer model of axis a,b,c and core radius eps
    and max radius of rmax.

    rho = (1.+(r/eps)**2)**(-5/2)

    """

    if isinstance(n, np.ndarray):
        random1 = n[:, 0]
        random2 = n[:, 1]
        random3 = n[:, 2]
        n = len(n)
        ntot = mpi.mpi_allreduce(n)
    else:

        # set random seed
        np.random.seed(irand + mpi.mpi_Rank())
        # set the number of particles per procs
        n, ntot = get_local_n(n)

        random1 = np.random.random([n])
        random2 = np.random.random([n])
        random3 = np.random.random([n])

    # positions
    rmax = float(rmax)
    eps = float(eps)

    eps = eps / rmax
    xm = 1. / (1. + (eps)**2) * random1**(2. / 3.)
    xm = eps * np.sqrt(xm / (1. - xm))
    phi = 2 * np.pi * random2

    costh = 1. - 2. * random3
    sinth = np.sqrt(1. - costh**2)
    axm = rmax * a * xm * sinth
    bxm = rmax * b * xm * sinth
    x = axm * np.cos(phi)
    y = bxm * np.sin(phi)
    z = rmax * c * xm * costh

    pos = np.transpose(np.array([x, y, z]))
    # velocities
    if vel == 'yes':

        R = np.sqrt(x**2 + y**2)
        rho = (3. * M / (4. * np.pi * eps**3)) * \
            (1 + (R**2 + z**2) / eps**2)**(-5. / 2.)
        C2 = z**2 + eps**2
        C = np.sqrt(C2)

        TD = M * C / (R**2 + C2)**(3. / 2.)
        sz = np.sqrt(eps**2 / (8. * np.pi * C2) / rho) * TD

        vx = sz * np.random.standard_normal([n])
        vy = sz * np.random.standard_normal([n])
        vz = sz * np.random.standard_normal([n])
        vel = np.transpose(np.array([vx, vy, vz]))

    else:
        vel = np.ones([n, 3]) * 0.0

    # masses
    mass = np.ones([n]) * M / ntot

    nb = Nbody(
        status='new',
        p_name=name,
        pos=pos,
        vel=vel,
        mass=mass,
        ftype=ftype)

    return nb


def kuzmin(
        n,
        eps,
        dz,
        irand=1,
        name='kuzmin.hdf5',
        ftype='swift'):
    """
    Return an Nbody object that contains n particles distributed
    in a Kuzmin (infinitely thin) disk

    rho = eps*M/(2*pi*(R**2+eps**2)**(3/2))

    """

    if isinstance(n, np.ndarray):
        random1 = n[:, 0]
        random2 = n[:, 1]
        random3 = n[:, 2]
        n = len(n)
        ntot = mpi.mpi_allreduce(n)
    else:

        # set random seed
        np.random.seed(irand + mpi.mpi_Rank())
        # set the number of particles per procs
        n, ntot = get_local_n(n)

        random1 = np.random.random([n])
        random2 = np.random.random([n])
        random3 = np.random.random([n])

    rmax = 1
    xx = random1
    xx = np.sqrt((eps / (1 - xx))**2 - eps**2)
    theta = 2. * random2 * np.pi

    x = rmax * xx * np.cos(theta)
    y = rmax * xx * np.sin(theta)
    z = dz * random3 - dz / 2.

    pos = np.transpose(np.array([x, y, z]))
    vel = np.ones([n, 3]) * 0.0
    mass = np.ones([n]) * 1. / ntot

    nb = Nbody(
        status='new',
        p_name=name,
        pos=pos,
        vel=vel,
        mass=mass,
        ftype=ftype)

    return nb


def expd(
        n,
        Hr,
        Hz,
        Rmax,
        Zmax,
        irand=1,
        name='expd.hdf5',
        ftype='swift'):
    """
    Exonential disk

    rho = 1/(1+(r/rc)**2)
    """

    # set random seed
    irand = irand + mpi.mpi_Rank()
    # set the number of particles per procs
    n, ntot = get_local_n(n)

    pos = iclib.exponential_disk(n, Hr, Hz, Rmax, Zmax, irand)
    vel = np.ones([n, 3]) * 0.0
    mass = np.ones([n]) * 1. / ntot

    nb = Nbody(
        status='new',
        p_name=name,
        pos=pos,
        vel=vel,
        mass=mass,
        ftype=ftype)

    return nb


def miyamoto_nagai(
    n, a, b, Rmax,
    Zmax, irand=1,
    fct=None,
    fRmax=0,
    name='miyamoto.hdf5',
    ftype='swift'):
    """
    Miyamoto Nagai distribution
    """

    # set random seed
    irand = irand + mpi.mpi_Rank()
    # set the number of particles per procs
    n, ntot = get_local_n(n)

    if fct is None:
        pos = iclib.miyamoto_nagai(n, a, b, Rmax, Zmax, irand)
    else:
        pos = iclib.miyamoto_nagai_f(n, a, b, Rmax, Zmax, irand, fct, fRmax)

    vel = np.ones([n, 3]) * 0.0
    mass = np.ones([n]) * 1. / ntot

    nb = Nbody(
        status='new',
        p_name=name,
        pos=pos,
        vel=vel,
        mass=mass,
        ftype=ftype)

    return nb


def generic_alpha(
        n,
        a,
        e,
        rmax,
        irand=1,
        fct=None,
        name='generic_alpha.hdf5',
        ftype='swift'):
    """
    Generic alpha distribution : rho ~ (r+e)^a
    """

    # set random seed
    irand = irand + mpi.mpi_Rank()
    # set the number of particles per procs
    n, ntot = get_local_n(n)

    pos = iclib.generic_alpha(n, a, e, rmax, irand)

    vel = np.ones([n, 3]) * 0.0
    mass = np.ones([n]) * 1. / ntot

    nb = Nbody(
        status='new',
        p_name=name,
        pos=pos,
        vel=vel,
        mass=mass,
        ftype=ftype)

    return nb


def dl2_mr(r, args):
    """
    Mass in the radius r for the distribution

    rho = (1.-eps*(r/rmax)**2)
    """
    eps = args[0]
    rmax = args[1]
    return ((4. / 3.) * r**3 - (4. / 5.) * eps * r**5 / rmax**2) / \
        (((4. / 3.) - (4. / 5.) * eps) * rmax**3)


def dl2(
    n,
    a,
    b,
    c,
    eps,
    rmax,
    irand=1,
    name='dl2.hdf5',
    ftype='swift'):
    """
    Return an Nbody object that contains n particles distributed as

    rho = (1.-eps*(r/rmax)**2)
    """

    if isinstance(n, np.ndarray):
        random1 = n[:, 0]
        random2 = n[:, 1]
        random3 = n[:, 2]
        n = len(n)
        ntot = mpi.mpi_allreduce(n)
    else:

        # set random seed
        np.random.seed(irand + mpi.mpi_Rank())
        # set the number of particles per procs
        n, ntot = get_local_n(n)

        random1 = np.random.random([n])
        random2 = np.random.random([n])
        random3 = np.random.random([n])

    x = random1
    xm = invert(x, 0, rmax, dl2_mr, [eps, rmax])
    phi = 2 * np.pi * random2

    costh = 1. - 2. * random3
    sinth = np.sqrt(1. - costh**2)
    axm = a * xm * sinth
    bxm = b * xm * sinth
    x = axm * np.cos(phi)
    y = bxm * np.sin(phi)
    z = c * xm * costh

    pos = np.transpose(np.array([x, y, z]))
    vel = np.ones([n, 3]) * 0.0
    mass = np.ones([n]) * 1. / ntot

    nb = Nbody(
        status='new',
        p_name=name,
        pos=pos,
        vel=vel,
        mass=mass,
        ftype=ftype)

    return nb


def isothm_mr(r, args):
    """
    Mass in the radius r for the distribution

    rho = 1/(1+r/rc)**2
    """
    rc = args[0]
    rm = args[1]

    cte = 2 * rc**3 * np.log(rc) + rc**3

    Mr = r * rc**2 - 2 * rc**3 * np.log(rc + r) - rc**4 / (rc + r) + cte
    Mx = rm * rc**2 - 2 * rc**3 * np.log(rc + rm) - rc**4 / (rc + rm) + cte

    return Mr / Mx


def isothm(
        n,
        rc,
        rmax,
        irand=1,
        name='isothm.hdf5',
        ftype='swift'):
    """
    Return an Nbody object that contains n particles distributed as

    rho = 1/(1+r/rc)**2
    """

    if isinstance(n, np.ndarray):
        random1 = n[:, 0]
        random2 = n[:, 1]
        random3 = n[:, 2]
        n = len(n)
        ntot = mpi.mpi_allreduce(n)
    else:

        # set random seed
        np.random.seed(irand + mpi.mpi_Rank())
        # set the number of particles per procs
        n, ntot = get_local_n(n)

        random1 = np.random.random([n])
        random2 = np.random.random([n])
        random3 = np.random.random([n])

    x = random1
    xm = invert(x, 0, rmax, isothm_mr, [rc, rmax])
    phi = 2 * np.pi * random2

    costh = 1. - 2. * random3
    sinth = np.sqrt(1. - costh**2)
    axm = xm * sinth
    bxm = xm * sinth
    x = axm * np.cos(phi)
    y = bxm * np.sin(phi)
    z = xm * costh

    pos = np.transpose(np.array([x, y, z]))
    vel = np.ones([n, 3]) * 0.0
    mass = np.ones([n]) * 1. / ntot

    nb = Nbody(
        status='new',
        p_name=name,
        pos=pos,
        vel=vel,
        mass=mass,
        ftype=ftype)

    return nb


def pisothm_mr(r, args):
    """
    Mass in the radius r for the distribution

    rho = 1/(1+(r/rc)**2)
    """
    rc = args[0]
    rmn = args[1]
    rmx = args[2]

    Mr = rc**3 * (r / rc - np.arctan(r / rc))
    Mmn = rc**3 * (rmn / rc - np.arctan(rmn / rc))
    Mmx = rc**3 * (rmx / rc - np.arctan(rmx / rc))

    return (Mr - Mmn) / (Mmx - Mmn)


def pisothm(
        n,
        rc,
        rmax,
        rmin=0,
        irand=1,
        name='pisothm.hdf5',
        ftype='swift'):
    """
    Pseudo-isothermal sphere
    Mass in the radius r for the distribution

    rho = 1/(1+(r/rc)**2)
    """

    if isinstance(n, np.ndarray):
        random1 = n[:, 0]
        random2 = n[:, 1]
        random3 = n[:, 2]
        n = len(n)
        ntot = mpi.mpi_allreduce(n)
    else:

        # set random seed
        np.random.seed(irand + mpi.mpi_Rank())
        # set the number of particles per procs
        n, ntot = get_local_n(n)

        random1 = np.random.random([n])
        random2 = np.random.random([n])
        random3 = np.random.random([n])

    x = random1
    xm = invert(x, rmin, rmax, pisothm_mr, [rc, rmin, rmax])
    phi = 2 * np.pi * random2

    costh = 1. - 2. * random3
    sinth = np.sqrt(1. - costh**2)
    axm = xm * sinth
    bxm = xm * sinth
    x = axm * np.cos(phi)
    y = bxm * np.sin(phi)
    z = xm * costh

    pos = np.transpose(np.array([x, y, z]))
    vel = np.ones([n, 3]) * 0.0
    mass = np.ones([n]) * 1. / ntot

    nb = Nbody(
        status='new',
        p_name=name,
        pos=pos,
        vel=vel,
        mass=mass,
        ftype=ftype)

    return nb


def nfw(
        n,
        rs,
        Rmax,
        dR,
        Rs=None,
        irand=1,
        name='nfw.hdf5',
        ftype='swift',
        verbose=False):
    """
    Return an Nbody object that contains n particles following
    an nfw profile.

    rho = 1/[ (r/rs)(1+r/rs)^2 ]

    """

    def Mr(r, rs):
        return 4 * np.pi * rs**3 * (np.log(1. + r / rs) - r / (rs + r))

    if not isinstance(Rs, np.ndarray):
        Rs = np.arange(0, Rmax + dR, dR)		  # should use a non linear grid

    Mrs = np.zeros(len(Rs))
    ntot = len(Rs)

    for i in range(len(Rs)):
        Mrs[i] = Mr(Rs[i], rs)
        if verbose:
            print((Rs[i], Mrs[i], i, '/', ntot))

    # normalisation
    Mrs = Mrs / Mrs[-1]
    Mrs[0] = 0

    # now use Mr
    nb = generic_Mr(
        n,
        rmax=Rmax,
        R=Rs,
        Mr=Mrs,
        irand=irand,
        name=name,
        ftype=ftype,
        verbose=verbose)

    return nb


def hernquist(
        n,
        rs,
        Rmax,
        dR,
        Rs=None,
        irand=1,
        name='hernquist.hdf5',
        ftype='swift',
        verbose=False):
    """
    Return an Nbody object that contains n particles following
    a hernquist modifed profile.

    rho =  1/( (r/rs) * (1+r/rs)**3 )

    """

    def Mr(r, rs):
        return rs**3 * 0.5 * (r / rs)**2 / (1 + r / rs)**2

    if not isinstance(Rs, np.ndarray):
        Rs = np.arange(0, Rmax + dR, dR)		  # should use a non linear grid

    Mrs = np.zeros(len(Rs))
    ntot = len(Rs)

    for i in range(len(Rs)):
        Mrs[i] = Mr(Rs[i], rs)
        if verbose:
            print((Rs[i], Mrs[i], i, '/', ntot))

    # normalisation
    Mrs = Mrs / Mrs[-1]
    Mrs[0] = 0

    # now use Mr
    nb = generic_Mr(
        n,
        rmax=Rmax,
        R=Rs,
        Mr=Mrs,
        irand=irand,
        name=name,
        ftype=ftype,
        verbose=verbose)

    return nb


def burkert(
        n,
        rs,
        Rmax,
        dR,
        Rs=None,
        irand=1,
        name='burkert.hdf5',
        ftype='swift',
        verbose=False):
    """
    Return an Nbody object that contains n particles following
    a burkert profile.

    rhob = 1 / ( ( 1 + r/rs  ) * ( 1 + (r/rs)**2  ) )

    """

    def Mr(r, rs):
        return 4 * np.pi * rs**3 * \
            (0.25 * np.log((r / rs)**2 + 1) - 0.5 * np.arctan(r / rs) + 0.5 * np.log((r / rs) + 1))

    if not isinstance(Rs, np.ndarray):
        Rs = np.arange(0, Rmax + dR, dR)		  # should use a non linear grid

    Mrs = np.zeros(len(Rs))
    ntot = len(Rs)

    for i in range(len(Rs)):
        Mrs[i] = Mr(Rs[i], rs)
        if verbose:
            print((Rs[i], Mrs[i], i, '/', ntot))

    # normalisation
    Mrs = Mrs / Mrs[-1]
    Mrs[0] = 0

    # now use Mr
    nb = generic_Mr(
        n,
        rmax=Rmax,
        R=Rs,
        Mr=Mrs,
        irand=irand,
        name=name,
        ftype=ftype,
        verbose=verbose)

    return nb


def nfwg(
        n,
        rs,
        gamma,
        Rmax,
        dR,
        Rs=None,
        irand=1,
        name='nfwg.hdf5',
        ftype='swift',
        verbose=False):
    """
    Return an Nbody object that contains n particles following
    an nfw modifed profile.

    rho = 1/[ ((r/rs))**(gamma)(1+r/rs)^2 ]**(0.5*(3-gamma))

    """

    if not is_scipy:
        raise Exception("module scipy needed for function nfwg !")

    def Mr(r, rs, gamma):
        aa = 1.5 - 0.5 * gamma
        cc = 2.5 - 0.5 * gamma
        z = -r**2 / rs**2
        return 2 * np.pi * (r / rs)**-gamma * r**3 * \
            special.hyp2f1(aa, aa, cc, z) / aa

    if not isinstance(Rs, np.ndarray):
        Rs = np.arange(0, Rmax + dR, dR)		  # should use a non linear grid

    Mrs = np.zeros(len(Rs))
    ntot = len(Rs)

    for i in range(len(Rs)):
        Mrs[i] = Mr(Rs[i], rs, gamma)
        if verbose:
            print((Rs[i], Mrs[i], i, '/', ntot))

    # normalisation
    Mrs = Mrs / Mrs[-1]
    Mrs[0] = 0

    # now use Mr
    nb = generic_Mr(
        n,
        rmax=Rmax,
        R=Rs,
        Mr=Mrs,
        irand=irand,
        name=name,
        ftype=ftype,
        verbose=verbose)

    return nb



def gen_2_slopes(
    n,
    rs,
    a,
    b,
    Rmax,
    dR,
    Rcut=None,
    power_cut=None,
    Rs=None,
    NR=1e4,
    irand=1,
    name='gen_2_slopes.hdf5',
    ftype='swift',
    verbose=False):
    """
    Return an Nbody object that contains n particles following
    an generic 2 slope profile.

    rho = 1/( (r/rs)**a * (1+r/rs)**(b-a) )

    Note : this is a re-implementation of generic2c

    """
        
    from scipy.integrate import quadrature, quad
    from .mass_models import gen_2_slopes

    def exp_cutoff(x, x_cut, speed):
      if x_cut is None or speed is None:
        return 1.
      else:  
        return np.exp( - (x / x_cut)**speed)
      

    if isinstance(n, np.ndarray):
        random1 = n[:, 0]
        random2 = n[:, 1]
        random3 = n[:, 2]
        n = len(n)
        ntot = mpi.mpi_allreduce(n)
    else:
        # set random seed
        np.random.seed(irand + mpi.mpi_Rank())
        # set the number of particles per procs
        n, ntot = get_local_n(n)
        random1 = np.random.random([n])
        random2 = np.random.random([n])
        random3 = np.random.random([n])


    gen2slopes = gen_2_slopes.GEN2SLOPES(alpha=a,beta=b,rho0=1,rs=rs,G=1)
    gen2slopes.info()
    
    # model density
    fctDensity_gen_2_slopes_cutoff = lambda x:gen2slopes.Density(x)*exp_cutoff(x, Rcut, power_cut)
    fctDensity = fctDensity_gen_2_slopes_cutoff
    fctMr = np.vectorize(lambda x: 4*np.pi *  quad( lambda r: r*r * fctDensity(r)/1 , 0., x)[0]  )
    Mmax = fctMr(Rmax)
    fctMr = np.vectorize(lambda x: 4*np.pi *  quad( lambda r: r*r * fctDensity(r)/Mmax , 0., x)[0]  )    

    if not isinstance(Rs, np.ndarray):      
      R = np.logspace(np.log10(dR),np.log10(Rmax),num=int(NR))   
    
    
    # cumulative mass  
    Mcum = fctMr(R)
    xm = np.interp(random1 ,Mcum ,R) 
    
    phi = random2 * np.pi * 2.
    costh = 1. - 2. * random3

    sinth = np.sqrt(1. - costh**2)
    axm = xm * sinth
    bxm = xm * sinth
    x = axm * np.cos(phi)
    y = bxm * np.sin(phi)
    z = xm * costh

    pos = np.transpose(np.array([x, y, z]))
    vel = np.ones([n, 3]) * 0.0
    mass = np.ones([n]) * 1. / ntot

    nb = Nbody(
        status='new',
        p_name=name,
        pos=pos,
        vel=vel,
        mass=mass,
        ftype=ftype)

    return nb



def generic2c(
    n,
    rs,
    a,
    b,
    Rmax,
    dR,
    Rs=None,
    irand=1,
    name='nfwg.hdf5',
    ftype='swift',
    verbose=False):
    """
    Return an Nbody object that contains n particles following
    an nfw modifed profile.

    rho = 1/( (r/rs)**a * (1+r/rs)**(b-a) )

    """

    if not is_scipy:
        raise Exception("module scipy needed for function generic2c !")

    def Mr(r, rs, a, b):
        a = float(a)
        b = float(b)

        aa = b - a
        bb = -a + 3
        cc = 4 - a
        z = -r / rs

        return 4 * np.pi * (r / rs)**(-a) * r**3 * \
            special.hyp2f1(aa, bb, cc, z) / bb

    if not isinstance(Rs, np.ndarray):
        Rs = np.arange(0, Rmax + dR, dR)		  # should use a non linear grid

    Mrs = np.zeros(len(Rs))
    ntot = len(Rs)

    for i in range(len(Rs)):
        Mrs[i] = Mr(Rs[i], rs, a, b)
        if verbose:
            print((Rs[i], Mrs[i], i, '/', ntot))

    # normalisation
    Mrs = Mrs / Mrs[-1]
    Mrs[0] = 0

    # now use Mr
    nb = generic_Mr(
        n,
        rmax=Rmax,
        R=Rs,
        Mr=Mrs,
        irand=irand,
        name=name,
        ftype=ftype,
        verbose=verbose)

    return nb



def deprojsersic(
    n,
    re,
    nindex,
    Rmax,
    M,
    dR,
    Rs=None,
    irand=1,
    name='nfwg.hdf5',
    ftype='swift',
    verbose=False):
    """
    Return an Nbody object that contains n particles following
    a 3D deprojected sersic profile.
    """
    
    from pNbody.mass_models import deprojsersic as sersic

    def Mr(r, re, nindex):
      return sersic.CumulativeMass(1, re, nindex, r)


    if not isinstance(Rs, np.ndarray):
        Rs = np.arange(0, Rmax + dR, dR)		  # should use a non linear grid

    Mrs = np.zeros(len(Rs))
    ntot = len(Rs)

    for i in range(len(Rs)):
        Mrs[i] = Mr(Rs[i], re, nindex)
        if verbose:
            print((Rs[i], Mrs[i], i, '/', ntot))

    # normalisation
    Mrs = Mrs / Mrs[-1]
    Mrs[0] = 0

    # now use Mr
    nb = generic_Mr(
        n,
        rmax=Rmax,
        R=Rs,
        Mr=Mrs,
        irand=irand,
        name=name,
        ftype=ftype,
        verbose=verbose)
    
    # set proper mass
    nb.mass = nb.mass/nb.mass.sum()*M 

    return nb


def power_spherical(
    n,
    alpha,
    r_c,
    amplitude,
    r_1,
    Rmax,
    dR,
    Rs=None,
    irand=1,
    name='psc.hdf5',
    ftype='swift',
    verbose=False):
    """
    Return an Nbody object that contains n particles following
    a power law with an exponentital cutoff following the density:
        \rho(r) = A (\frac{r_1}{r})^\alpha \exp(- \frac{r^2}{r_c^2})

    alpha  (alpha)     
    r_c    (r_c)        
    A      (amplitude)  
    
    """
    
    from pNbody.mass_models import powerSphericalCutoff as psc

    def Mr(r):
      return psc.CumulativeMass(alpha, r_c, amplitude, r_1, r)
          
    # total mass      
    M = Mr(Rmax)         
          
    if not isinstance(Rs, np.ndarray):
        Rs = np.arange(0, Rmax + dR, dR)		  # should use a non linear grid

    Mrs = np.zeros(len(Rs))
    ntot = len(Rs)

    for i in range(len(Rs)):
        Mrs[i] = Mr(Rs[i])
        if verbose:
            print((Rs[i], Mrs[i], i, '/', ntot))

    # normalisation
    Mrs = Mrs / Mrs[-1]
    Mrs[0] = 0
    
    
    # now use Mr
    nb = generic_Mr(n,rmax=Rmax,R=Rs,Mr=Mrs,irand=irand,name=name,ftype=ftype,verbose=verbose)
        
    # set proper mass
    nb.mass = nb.mass/nb.mass.sum()*M 

    return nb  
  
  



# def generic_MxHyHz(n,xmax,ymax,zmax,x=None,Mx=None,name='box_Mx.hdf5',ftype='swift',verbose=False):
#  """
#  Distribute particles in a box. The density in x is defined in order to reproduce M(x) given by Mx.
#  Here, contrary to generic_Mx, the size of the box is defined.
#  """
#
#  if type(nx) == np.ndarray:
#    random1 = nx
#    random2 = ny
#    random3 = nz
#    n = len(n)
#  else:
#    random1 = np.random.random([nx])
#    random2 = np.random.random([ny])
#    random3 = np.random.random([nz])
#
#  pos = iclib.generic_MxHyHz(n,xmax,ymax,zmax,x.astype(np.float32),Mx.astype(np.float32),random1.astype(np.float32),random2.astype(np.float32),random3.astype(np.float32),verbose)
#
#  vel =  np.ones([n,3])*0.0
#  mass = np.ones([n])*1./n
#
#  nb = Nbody(status='new',p_name=name,pos=pos,vel=vel,mass=mass,ftype=ftype)
#
#  return nb


#################################
# geometric distributions
#################################


def generic_Mx(
        n,
        xmax,
        x=None,
        Mx=None,
        irand=1,
        name='box_Mx.hdf5',
        ftype='swift',
        verbose=False):
    """
    Distribute particles in a box. The density in x is defined in order to reproduce M(x) given by Mx
    """

    if isinstance(n, np.ndarray):
        random1 = n[:, 0]
        random2 = n[:, 1]
        random3 = n[:, 2]
        n = len(n)
        ntot = mpi.mpi_allreduce(n)
    else:

        # set random seed
        np.random.seed(irand + mpi.mpi_Rank())
        # set the number of particles per procs
        n, ntot = get_local_n(n)

        random1 = np.random.random([n])
        random2 = np.random.random([n])
        random3 = np.random.random([n])

    pos = iclib.generic_Mx(
        n,
        xmax,
        x.astype(np.float32),
        Mx.astype(np.float32),
        random1.astype(np.float32),
        random2.astype(np.float32),
        random3.astype(np.float32),
        verbose)

    vel = np.ones([n, 3]) * 0.0
    mass = np.ones([n]) * 1. / ntot

    nb = Nbody(
        status='new',
        p_name=name,
        pos=pos,
        vel=vel,
        mass=mass,
        ftype=ftype,
        verbose=verbose)

    return nb


def generic_Mr(
        n,
        rmax,
        R=None,
        Mr=None,
        irand=1,
        name='sphere_Mr.hdf5',
        ftype='swift',
        verbose=False):
    """
    Distribute particles in order to reproduce M(R) given by Mr
    """

    if isinstance(n, np.ndarray):
        random1 = n[:, 0]
        random2 = n[:, 1]
        random3 = n[:, 2]
        n = len(n)
        ntot = mpi.mpi_allreduce(n)
    else:

        # set random seed
        np.random.seed(irand + mpi.mpi_Rank())
        # set the number of particles per procs
        n, ntot = get_local_n(n)

        random1 = np.random.random([n])
        random2 = np.random.random([n])
        random3 = np.random.random([n])

    pos = iclib.generic_Mr(
        n,
        rmax,
        R.astype(np.float32),
        Mr.astype(np.float32),
        random1.astype(np.float32),
        random2.astype(np.float32),
        random3.astype(np.float32),
        verbose)

    vel = np.ones([n, 3]) * 0.0
    mass = np.ones([n]) * 1. / ntot

    nb = Nbody(
        status='new',
        p_name=name,
        pos=pos,
        vel=vel,
        mass=mass,
        ftype=ftype,
        verbose=verbose)

    return nb


#################################
# geometric primitives
#################################


def line(M=1., name='line.hdf5', ftype='swift'):

    x = np.array([-0.5, 0.5])
    y = np.array([0, 0])
    z = np.array([0, 0])

    n = len(x)
    pos = np.transpose((x, y, z))
    mass = M * np.ones(n)
    nb = Nbody(status='new', pos=pos, mass=mass, p_name=name, ftype=ftype)
    return nb


def square(M=1., name='square.hdf5', ftype='swift'):

    x = np.array([-0.5, +0.5, +0.5, -0.5])
    y = np.array([-0.5, -0.5, +0.5, +0.5])
    z = np.array([0, 0, 0, 0])

    n = len(x)
    pos = np.transpose((x, y, z))
    mass = M * np.ones(n)
    nb = Nbody(status='new', pos=pos, mass=mass, p_name=name, ftype=ftype)
    return nb


def circle(n=10, M=1., name='circle.hdf5', ftype='swift'):

    t = np.arange(0, 2 * np.pi, 2 * np.pi / n)

    x = np.cos(t)
    y = np.sin(t)
    z = np.zeros(n)

    n = len(x)
    pos = np.transpose((x, y, z))
    mass = M * np.ones(n)
    nb = Nbody(status='new', pos=pos, mass=mass, p_name=name, ftype=ftype)
    return nb


def grid(n=10, m=10, M=1., name='grid.hdf5', ftype='swift'):

    dx = 1 / float(n)
    dy = 1 / float(m)

    xx = np.arange(0, 1 + dx, dx)
    yy = np.arange(0, 1 + dy, dy)

    x = np.zeros(4 * n * m)
    y = np.zeros(4 * n * m)
    z = np.zeros(4 * n * m)

    k = 0
    for i in range(n):
        for j in range(m):
            x[k + 0] = xx[i + 0] - 0.5
            y[k + 0] = yy[j + 0] - 0.5

            x[k + 1] = xx[i + 1] - 0.5
            y[k + 1] = yy[j + 0] - 0.5

            x[k + 2] = xx[i + 1] - 0.5
            y[k + 2] = yy[j + 1] - 0.5

            x[k + 3] = xx[i + 0] - 0.5
            y[k + 3] = yy[j + 1] - 0.5
            k = k + 4

    n = len(x)
    pos = np.transpose((x, y, z))
    mass = M * np.ones(n)
    nb = Nbody(status='new', pos=pos, mass=mass, p_name=name, ftype=ftype)
    return nb


def cube(M=1., name='cube.hdf5', ftype='swift'):
    x = np.array([0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1,
               0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, ]) - 0.5
    y = np.array([0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1,
               1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, ]) - 0.5
    z = np.array([0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1,
               1, 1, 1, 0, 1, 0, 1, 0, 1, 0, 1, ]) - 0.5

    n = len(x)
    pos = np.transpose((x, y, z))
    mass = M * np.ones(n)
    nb = Nbody(status='new', pos=pos, mass=mass, p_name=name, ftype=ftype)
    return nb


def sphere(n=10, m=10, M=1., name='sphere.hdf5', ftype='swift'):

    pos = np.zeros((2 * n * m, 3), np.float32)

    ts = np.arange(0, 2 * np.pi, 2 * np.pi / n)
    zs = 2 * np.arange(m) / float(m - 1) - 1

    k = 0

    # parallels
    for i in range(m):
        for j in range(n):
            r = np.sin(np.arccos(zs[i]))
            x = r * np.cos(ts[j])
            y = r * np.sin(ts[j])
            z = zs[i]

            pos[k] = [x, y, z]
            k = k + 1

    # meridians
    for j in range(n):
        for i in range(m):
            r = np.sin(np.arccos(zs[i]))
            x = r * np.cos(ts[j])
            y = r * np.sin(ts[j])
            z = zs[i]

            pos[k] = [x, y, z]
            k = k + 1

    nb = Nbody(status='new', pos=pos, p_name=name, ftype=ftype)
    nb.mass = M * nb.mass

    return nb


def arrow(M=1., name='arrow.hdf5', ftype='swift'):

    q = (1 + np.sqrt(5)) / 2.  # golden number

    lx = 1 / q
    x1 = lx / 3.
    x2 = 2 * lx / 3.
    y1 = 1. / q

    x = np.array([x1, x2, x2, lx, 0.5 * lx, 0, x1, x1])
    y = np.array([0, 0, y1, y1, 1, y1, y1, 0])
    z = np.zeros(len(x))

    n = len(x)
    pos = np.transpose((x, y, z))
    mass = M * np.ones(n)
    nb = Nbody(status='new', pos=pos, mass=mass, p_name=name, ftype=ftype)

    nb.translate([-lx / 2, -1, 0])
    nb.rotate(axis='z', angle=np.pi)

    return nb



#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      talkgdisp.py
#  brief:     talkgdisp.py
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################


import sys
import socket

###########################
class TalkServer:
    ###########################

    """
    talk server class
    """

    def __init__(self, address=None, module=None, cl=None):

        self.HOST = address['HOST']
        self.PORT = address['PORT']
        self.module = module
        self.cl = cl

    ###########################
    def isalive(self):
        ###########################

        # init the socket

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        PORTS = list(range(40000, 40016))

        for PORT in PORTS:
            sock.connect_ex((self.HOST, PORT))
            try:
                sock.send("isalive")
                rawanswer = sock.recv(128)
                sock.close()
                if rawanswer == 'isalive%&yes':
                    self.PORT = PORT
                    # print "cdteld is running on port",PORT
                    return 'yes'
            except BaseException:
                pass

        print((
            "socket error : verify that a server is running on host %s" %
            (self.HOST)))
        sys.exit()

    ###########################
    def talk(self, request, param):
        ###########################

        # init the socket

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect_ex((self.HOST, self.PORT))

        ######################
        # send the command
        ######################

        while True:

            try:
                sock.send(request + "&" + param)
                break
            except socket.error:
                sock.close()
                self.isalive()

        ######################
        # receive the answer
        ######################

        try:
            rawanswer = sock.recv(128)
            sock.close()
        except socket.error:
            sock.close()
            print("socket error : unable to receive the answer")
            rawanswer = -1

        ###################################
        # extract answer and check validity
        ###################################

        pos = rawanswer.find('%&')

        # check
        request_answ = rawanswer[:pos]
        print(request_answ)
        if request_answ != request + "&" + param:
            print(("error :", request_answ))
            return '-1'
        else:
            return rawanswer[pos + 2:]



#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      ic.py
#  brief:     generating initial conditions
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################


'''
This method is strongly inspired from the nbodpy code from Errani:
https://rerrani.github.io/code.html
https://github.com/rerrani/nbopy
'''


from pNbody import *
import numpy as np
from scipy.integrate import quad


class DistributionFunction():
  
  def __init__(self,Rmin=1e-2,Rmax=100,Rinf=np.inf,NR=1e4,NE=1e4,fctDensity=None,fctCumulativeMass=None,fctPotential=None,TotalMass=None,G=1.,EPSREL=1e-6):
          
    
    self.Rmin   = Rmin        # maximum sampled radius
    self.Rmax   = Rmax        # minimum sampled radius  
    self.Rinf   = Rinf        # maximal radius consdered to compute the total mass
    self.NR     = NR          # number of radius bins
    self.NE     = NE          # number of energy bins
    self.EPSREL = EPSREL      # integration relative precision  
    self.fctDensity        = fctDensity
    self.fctPotential      = fctPotential
    self.fctCumulativeMass = fctCumulativeMass
    self.TotalMass         = TotalMass
    self.G                 = G           # gravitational constant

    # initialize 
    self.Init()

    # compute Psi array
    self.setPsi()
    
    # compute Nu array  (normalized rho)
    self.setNu()

    # define the distribution function
    self.setDFfct()
    
    # define energy bins
    self.setEnergyBins()
    
    
    
  def Init(self):
    '''
    Perform initialization
    '''
    
    # set the radial bins
    self.setRadialBins()
    
    # a one variable function (of R)
    self.computeTotalMass()
    
    # define the cumulative mass function
    self.setCumulativeMassfct()
    
    # define the potential function
    self.setPotentialfct()
          
    # compute the cumulative mass
    self.computeCumulativeMass()
    
  
  def setRadialBins(self,log=True):  
    '''
    define a list of radius
    '''
    if log:
      self.R = np.logspace(np.log10(self.Rmin),np.log10(self.Rmax),num=int(self.NR))      
    else:
      self.R = np.linspace(self.Rmin,self.Rmax,num=int(self.NR))     

  def setEnergyBins(self,log=False):  
    '''
    define a list of energy, from max(Psi) (R=Rmin) to Emax/NE  (avoid E=0)
    '''
        
    # compute most-bound/least-bound energy
    self.Emax = self.psi[0]
    self.Emin = self.Emax/float(self.NE)
    
    if log:
      self.E = np.logspace(np.log10(self.Emin),np.log10(self.Emax),num=int(self.NE))      
    else:
      self.E = np.linspace(self.Emin,self.Emax,num=int(self.NE))    


  
  def computeTotalMass(self):
    ''' 
    compute the total mass function by numerical integration
    '''
    if self.TotalMass is None:
      self.Mtot =  4 * np.pi * quad( lambda r: r*r * self.fctDensity(r) , 0., self.Rinf)[0] 
    else:
      self.Mtot = self.TotalMass

  def setCumulativeMassfct(self):
    ''' 
    define the "normalized" cumulative mass function (direct integration)
    '''
    if self.fctCumulativeMass is None:
      self.fctMr = np.vectorize(lambda x: 4*np.pi *  quad( lambda r: r*r * self.fctDensity(r)/self.Mtot , 0., x)[0]  )
    else:
      self.fctMr = lambda r:self.fctCumulativeMass(r)/self.Mtot
        
  
  def computeCumulativeMass(self):
    '''
    compute the "normalized" cumulative mass
    '''     
    self.Mcum = self.fctMr(self.R)
    
    

        
  def setPotentialfct(self):
    ''' 
    define the potential function (it is faster to use quad instead of fctMr)
    '''
    
    if self.fctPotential is None:
       # self.fctPotential  = np.vectorize(lambda x: -4*np.pi*self.G*quad( lambda r: r*r * self.fctDensity(r)/self.Mtot , 0., x)[0]/x 
                                                   # -4*np.pi*self.G*quad( lambda r: r   * self.fctDensity(r)/self.Mtot , x, self.Rinf)[0])
       self.fctPotential  = np.vectorize(lambda x: -4*np.pi*self.G*quad( lambda r: r*r * self.fctDensity(r) , 0., x)[0]/x 
                                                   -4*np.pi*self.G*quad( lambda r: r  * self.fctDensity(r) , x, self.Rinf)[0])
        
  
  def setPsi(self):
    '''
    set Psi (array), the relative potential
    '''
    self.psi = -self.fctPotential(self.R)
  
  def setNu(self):
    '''
    set Nu (array), the probability (normalized density)
    '''
    self.nu  =   self.fctDensity(self.R) / self.Mtot
      
    
    
  def setDFfct(self):
    ''' 
    define the distribution function through the Eddington formula.
    Note that we assume dnu/dpsi equal to 0.
    '''
      
    # compute gradients
    dndp   = np.gradient(self.nu,   self.psi)
    d2nd2p = np.gradient(dndp, self.psi)
    
    self.dndp   = dndp
    self.d2nd2p = d2nd2p
        
    if np.fabs(dndp[-1]) > 1e-5:
      print("WARNING: dnu/dpsi = %g > 1e-6\n"%dndp[-1])
      exit()

    # this fuction is very steep : we should integrate its log    
    #fctp =  lambda p:  np.interp(p, self.psi[::-1], d2nd2p[::-1]) / np.sqrt(e-p)
    self.fctDF = np.vectorize( lambda e: 1./(np.sqrt(8)*np.pi*np.pi) * (quad( lambda p:  np.interp(p, self.psi[::-1], d2nd2p[::-1]) / np.sqrt(e-p) , 0., e,  epsrel=self.EPSREL)[0]  ) )
    
     
  def computeDF(self):
    '''
    Compute the DF using the set of relative energies self.E
    '''
    self.DF = self.fctDF(self.E)
    
    
 
  def checkDF(self):
    '''
    check whether DF is physical - if not, check for rounding errors in integration, try increasing/decreasing NR, NE
    '''
    if np.any (self.DF < 0) : 
      print("DF < 0 !!!" )
      sys.exit(0)
    
    return True
    

  def clean(self):
    '''
    clean DF
    '''
    
    if np.any (self.DF < 0) : 
      pass
    else:
      return
     
    c = self.DF < 0
    
    print("INFO: {0:d} energy bins out of {1:d} have negative values".format(c.sum(),len(self.DF)))
    
    subDF = np.compress(c,self.DF)
    subE  = np.compress(c,self.E)
    
    self.DF = np.where(c,0,self.DF)

    
    
    
    
  def dPdr(self,e,r):  
    '''
    phase space volume element per energy accessible at given energy e and fixed radius r
    sqrt( 2(psi-e) ) r^2
    see BT, integrand of Eq 4.56
    '''  
    return np.sqrt( 2.*(np.interp(r, self.R, self.psi) - e )) * r*r     
    

  def PLikelihood(self,e,r):
    '''
    likelihood of a particle to have an energy E at a fixed radius
    dP/dr * DF(e)
    '''
    return np.interp(e,self.E,self.DF) * self.dPdr(e,r)
  
  
  def computePLikelihoodForR(self,R):
    '''
    compute the energy probability distribution 
    for a given radius
    '''

    # compute Psi for a given radius
    Epsi = np.interp(R,self.R,self.psi)
    
    # keep only the accessible energies for this radius
    #allowed = np.where (self.E <=  Epsi )[0]
    condition = self.E <=  Epsi
    Eallowed = np.compress (condition, self.E)
    
    # compute the maximum likelihood for this radius
    return self.PLikelihood(Eallowed,R),Eallowed
    

  def computeMaxLikelihood(self):
    '''
    for each radius in self.R, compute the maximum likelihood
    '''
    self.maxPLikelihood = np.zeros(int(self.NR))
    # loop over radii
    for i,R in enumerate(self.R):
      # compute the maximum likelihood for this radius
      PLikelihood,E = self.computePLikelihoodForR(R)
      
      # compute the maximum likelihood for this radius
      maxPLikelihood = PLikelihood.max()

      # and add 10% tolerance
      self.maxPLikelihood[i]=1.1* maxPLikelihood
    
    
  def sample(self,N,Ndraw,irand=0):
    '''
    N      : number of particles to draw
    Ndraw  : number of random numbers drawn at a time 
    
    return pos and vel
    '''
    np.random.seed(irand)
    
    xx = np.zeros(int(N))
    yy = np.zeros(int(N))
    zz = np.zeros(int(N))
    vx = np.zeros(int(N))
    vy = np.zeros(int(N))
    vz = np.zeros(int(N))    
    
    n=0             # current number of generated 'particles'
    Efails = 0      # rejection sampling failures in Energy

    # "normalized" cumulative mass
    Mcum = self.Mcum
    
    # compute the number of particles inside Rmin
    Nin  = int(N *      4 * np.pi * quad( lambda r: r*r * self.fctDensity(r)/self.Mtot , 0., self.Rmin)[0] )
    # compute the number of particles outside Rmax
    Nout = int(N * (1 - 4 * np.pi * quad( lambda r: r*r * self.fctDensity(r)/self.Mtot , 0., self.Rmax)[0] ) )
    
    if Nin > 0:
      print("INFO: out of %i particles there are %i part. with R < Rmin (%.2f per cent)"%(N,Nin, 100.*Nin/float(N) ) )
    if Nout > 0:
      print("INFO: out of %i particles there are %i part. with R > Rmax (%.2f per cent)"%(N,Nout,100.*Nout/float(N) ) )
        

    # while we still need to generate 'particles'..
    while (n < N):

      # random number between fraction of particles within  min. sampled radius Rmin 
      #                   and fraction of particles outside max. sampled radius Rmax
      randMcum = Nin/float(N)  + (1.-(Nin+Nout)/float(N)) * np.random.rand(int(Ndraw))

      # inverse the cumulative mass function to get a list of radius between Rmin and Rmax 
      # that sample the density rho(r)
      randR = np.interp(randMcum ,Mcum ,self.R)      

      
      # compute the relative potential
      psiR = np.interp(randR ,self.R ,self.psi)
      
      # random relative energy between 0 and psiR
      randE = np.random.rand(int(Ndraw)) * psiR  
      
      # likelihood for E at given R
      rhoE  = self.PLikelihood(randE,randR)
      
      # draw a random number between 0 and the max likelihood for each radius
      randY = np.random.rand(int(Ndraw)) * np.interp(randR,self.R,self.maxPLikelihood) 
      
      # record the index of the failed points
      Missidx = np.where(randY > rhoE)[0]        
      Efails += len(Missidx)
      
      #print("Efails = %d"%Efails)

      # repeat sampling at fixed R till we got all the energies we need
      loop=0
      while len(Missidx):
        loop=loop+1
        randE[Missidx] = np.random.rand(len(Missidx)) * psiR[Missidx]  
        rhoE[Missidx]  = self.PLikelihood(randE[Missidx],randR[Missidx])
        randY[Missidx] = np.random.rand(len(Missidx)) * np.interp(randR[Missidx],self.R, self.maxPLikelihood)
        Missidx = np.where(randY > rhoE)[0]
        Efails += len(Missidx)
        
        #print("loop=%3d Efails = %d"%(loop,Efails))
      

      # final check (this should never happen)
      okEidx = np.where(randY <= rhoE)[0]
      if len(okEidx) != int(Ndraw):             
        print("Particles went missing. Exit." )
        sys.exit(0)

      # Let's select as many R,E combinations as we're still missing to get N particles in total
      missing = int(N) - int(n)
      if len(okEidx) <= missing: 
        arraxIdx = n + np.arange(0,len(okEidx))
      else: 
        arraxIdx = n + np.arange(0,missing)
        okEidx = okEidx[:missing]      
      
      n += len(okEidx)


      # spherical symmetric model, draw random points on sphere
      Rtheta  = np.arccos (2. * np.random.rand(len(okEidx)) - 1.)
      Rphi    = np.random.rand(len(okEidx)) * 2*np.pi
      
      # isotropic velocity dispersion, draw random points on sphere
      Vtheta  = np.arccos (2. * np.random.rand(len(okEidx)) - 1.)
      Vphi    = np.random.rand(len(okEidx)) * 2*np.pi
      V =  np.sqrt( 2.*(psiR[okEidx] - randE[okEidx] ))  
      
      # spherical to cartesian coordinates 
      xx[arraxIdx] = randR[okEidx] * np.sin(Rtheta) * np.cos(Rphi)
      yy[arraxIdx] = randR[okEidx] * np.sin(Rtheta) * np.sin(Rphi)
      zz[arraxIdx] = randR[okEidx] * np.cos(Rtheta) 
      
      vx[arraxIdx] = V * np.sin(Vtheta) * np.cos(Vphi)
      vy[arraxIdx] = V * np.sin(Vtheta) * np.sin(Vphi)
      vz[arraxIdx] = V * np.cos(Vtheta)

    self.pos = np.transpose(np.array([xx, yy, zz]))
    self.vel = np.transpose(np.array([vx, vy, vz]))

    M = self.Mtot
    self.mass = self.Mtot*np.ones(len(self.pos))/len(self.pos)    




  def save(self,name,ftype,ptype,u):   
    '''
    save the model
    '''
    if name is not None:
      M = self.Mtot
            
      mass = M*np.ones(len(self.pos))/len(self.pos)
      nb = Nbody(status='new',p_name=name,pos=self.pos,vel=self.vel,mass=mass,ftype=ftype)
      
      # add units
      nb.UnitLength_in_cm         = u["UnitLength_in_cm"]
      nb.UnitMass_in_g            = u["UnitMass_in_g"]           
      nb.UnitVelocity_in_cm_per_s = u["UnitVelocity_in_cm_per_s"]
      nb.Unit_time_in_cgs         = u["Unit_time_in_cgs"]
    
      # additional stuffs
      nb.massarr=None
      nb.nzero = None
      nb.setComovingIntegrationOff()
      
      # particle type
      nb.set_tpe(ptype)
      
      # cvcenter
      nb.cvcenter()
      
      # write
      nb.write()
    
    
    
    
    
    
    




#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      pseudoisothermal.py
#  brief:     Defines pseudo-isothermal profiles
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################


"""
pseudo-isothermal model
"""

import numpy as np


def Potential(rho0, a, r, G=1.):
    """
    Potential
    """
    return 4*np.pi*G*rho0*a**2*(0.5*np.log(a**2+r**2) + a/r * np.arctan(r/a))


def Density(rho0, a, r, G=1.):
    """
    Density
    """
    return rho0 * 1/(1+(r**2/a**2))
    


def Vcirc(rho0, a, r, G=1.):
    """
    circular velocity
    """
        
    return np.sqrt(4*np.pi*G*rho0*a**2 * (1-a/r * np.arctan(r/a)))
    
    



###########################################################################################
#  package:   pNbody
#  file:      swift_logger.py
#  brief:
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>, Loic Hausammann <loic_hausammann@hotmail.com>
#
# This file is part of pNbody.
###########################################################################################


##########################################################################
#
# SWIFT LOGGER CLASS
#
##########################################################################

import sys
import types
import string
import numpy as np
import h5py
from copy import deepcopy

import pNbody
from pNbody import mpi, error, units, thermodyn
from astropy import units
try:				# all this is useful to read files
    from mpi4py import MPI
except BaseException:
    MPI = None

class Nbody_logger:

    def _init_spec(self):
        return

    def get_excluded_extension(self):
        """
        Return a list of file to avoid when extending the default class.
        """
        return []

    def check_spec_ftype(self):
        try:
            import h5py
            if mpi.NTask > 1:
                fd = h5py.File(
                    self.p_name_global[0],
                    'r',
                    driver="mpio",
                    comm=MPI.COMM_WORLD)
            else:
                fd = h5py.File(self.p_name_global[0],'r')
            test = "PartType0/Offset" not in fd
            fd.close()
            if test:
                raise error.FormatError("logger")
        except IOError as e:
            self.warning("logger not recognized: %s" % e)
            raise error.FormatError("logger")

    def get_read_fcts(self):
        return [self.read_particles]

    def get_write_fcts(self):
        return [self.write_particles]

    def get_mxntpe(self):
        return 6

    def get_header_translation(self):
        """
        Gives a dictionnary containing all the header translation.
        If a new variable is possible in the HDF5 format, only the translation
        is required for the reader/writer.
        As h5py is not supporting dictionnary, they need special care when reading/writing.
        """

    def get_default_spec_vars(self):
        """
        return specific variables default values for the class
        """

        return {'massarr': np.array([0, 0, self.nbody, 0, 0, 0]),
                'atime': 0.,
                'redshift': 0.,
                'flag_sfr': 0,
                'flag_feedback': 0,
                'nall': np.array([0, 0, self.nbody, 0, 0, 0]),
                'npart': np.array([0, 0, self.nbody, 0, 0, 0]),
                'flag_cooling': 0,
                'num_files': 1,
                'boxsize': 0.,
                'omega0': 0.,
                'omegalambda': 0.,
                'hubbleparam': 0.,
                'flag_age': 0.,
                'hubbleparam': 0.,
                'flag_metals': 0.,
                'nallhw': np.array([0, 0, 0, 0, 0, 0]),
                'flag_entr_ic': 0,
                'flag_chimie_extraheader': 0,
                'critical_energy_spec': 0.,
                'empty': 48 * '',
                'comovingintegration': None,
                'ChimieNelements': 0,
                }

    def get_massarr_and_nzero(self):
        """
        return massarr and nzero

        !!! when used in //, if only a proc has a star particle,
        !!! nzero is set to 1 for all cpu, while massarr has a length of zero !!!
        """

    def getHeaderTranslation(self):
        header_var = {}

        # Size variables
        header_var["Header/NumPart_ThisFile"] = "npart"
        header_var["Header/NumPart_Total"] = "nall"
        header_var["Header/NumPart_Total_HighWord"] = "nallhw"
        header_var["Header/MassTable"] = "massarr"
        header_var["Header/NumFilesPerSnapshot"] = "num_files"
        header_var["Header/BoxSize"] = "boxsize"
        header_var["Header/Time Offset"] = "time_offset"

        # Physics
        header_var["Header/Time"] = "atime"

        # Units
        header_var["Units/Unit length in cgs (U_L)"] = "UnitLength_in_cm"
        header_var["Units/Unit mass in cgs (U_M)"] = "UnitMass_in_g"
        header_var["Units/Unit time in cgs (U_t)"] = "Unit_time_in_cgs"
        header_var["Units/Unit temperature in cgs (U_T)"] = "Unit_temp_in_cgs"
        header_var["Units/Unit current in cgs (U_I)"] = "Unit_current_in_cgs"

        # Code
        header_var["Code/Code"] = "Unit_current_in_cgs"
        header_var["Code/CFLAGS"] = "cflags"
        header_var["Code/Code Version"] = "code_version"
        header_var["Code/Compiler Name"] = "compiler_name"
        header_var["Code/Compiler Version"] = "compiler_version"
        header_var["Code/Configuration options"] = "config_options"
        header_var["Code/FFTW library version"] = "fftw_lib_version"
        header_var["Code/Git Branch"] = "git_branch"
        header_var["Code/Git Date"] = "git_date"
        header_var["Code/Git Revision"] = "git_revision"
        header_var["Code/HDF5 library version"] = "hdf5_lib_version"
        header_var["Code/MPI library"] = "mpi_lib"

        # HydroScheme
        header_var["HydroScheme/Adiabatic index"] = "adiabatic_index"
        header_var["HydroScheme/CFL parameter"] = "cfl_parameter"
        header_var["HydroScheme/Dimension"] = "dimension"
        header_var["HydroScheme/Kernel delta N_ngb"] = "kernel_delta_n_ngb"
        header_var["HydroScheme/Kernel eta"] = "kernel_eta"
        header_var["HydroScheme/Kernel function"] = "kernel_function"
        header_var["HydroScheme/Kernel target N_ngb"] = "kernel_target_n_ngb"
        header_var["HydroScheme/Max ghost iterations"] = "max_ghost_iterations"
        header_var["HydroScheme/Maximal smoothing length"] = "maximal_smoothing_length"
        header_var["HydroScheme/Scheme"] = "scheme"
        header_var["HydroScheme/Smoothing length tolerance"] = "smoothing_length_tolerance"
        header_var["HydroScheme/Thermal Conductivity Model"] = "thermal_conductivity_model"
        header_var["HydroScheme/Viscosity Model"] = "viscosity_model"
        header_var["HydroScheme/Viscosity alpha"] = "viscosity_alpha"
        header_var["HydroScheme/Viscosity beta"] = "viscosity_beta"
        header_var["HydroScheme/Volume log(max(delta h))"] = "volume_log"
        header_var["HydroScheme/Volume max change time-step"] = "volume_max_change"

        # Swift directory
        header_var["RuntimePars/PeriodicBoundariesOn"] = "periodic"

        return header_var

    def getArrayTranslation(self):
        array_var = {}

        array_var["ParticleIDs"] = "num"
        array_var["Offset"] = "offset"

        return array_var

    def getArrayDefaultValue(self):
        """
        Gives a dictionary of default value for pNbody's arrays
        """
        # default value
        dval = {}
        dval["ParticleIDs"] = -1
        dval["Offset"] = -1
        return dval

    def getArrayDimension(self):
        """
        Gives a dictionary of dimension for pNbody's arrays
        """
        # dimension
        vdim = {}
        vdim["ParticleIDs"] = 1
        vdim["Offset"] = 1
        return vdim

    def get_list_header(self):
        """
        Gives a list of header directory from self.get_header_translation
        """
        list_header = []
        trans = self.getHeaderTranslation()
        for key, tmp in list(trans.items()):
            directory = key.split("/")[0]
            if directory not in list_header:
                list_header.append(directory)
        return list_header

    def read_particles(self, f):
        """
        read logger file
        """
        import libswiftlogger as logger

        if f is not None:
            f.seek(0, 2)

        fd = h5py.File(self.p_name_global[0], 'r')

        trans = self.getHeaderTranslation()
        list_header = self.get_list_header()

        for name in list_header:
            # e.g. create self.npart with the value
            # fd["Header"].attrs["NumPart_ThisFile"]
            for key in fd[name].attrs:
                full_name = name + "/" + key
                if full_name not in list(trans.keys()):
                    self.warning(
                        "'%s' key not recognized, please add it to config/format/swift_logger.py file in get_header_translation" %
                        full_name)
                else:
                    setattr(self, trans[full_name], fd[name].attrs[key])

        n = np.sum(self.npart)

        list_of_vectors = []

        ntab = self.getArrayTranslation()
        vdim = self.getArrayDimension()
        dval = self.getArrayDefaultValue()

        for i, n in enumerate(self.npart):

            if n == 0:
                continue

            if list(fd.keys()).count("PartType%d" % i) == 0:
                self.warning(
                    "type=%d n=%d but group %s is not found !" %
                    (i, n, "PartType%d" %
                     i))
                sys.exit()

            # loop over dataset
            block = fd["PartType%d" % i]
            for key in list(block.keys()):
                if key not in ntab:
                    self.warning(
                        "get a dataset with name %s but no such key in ntab" %
                        key)
                    continue

                varname = ntab[key]

                if not self.has_var(varname):
                    setattr(self, varname, None)

                # record the variable in a list
                if list_of_vectors.count(varname) == 0:
                    list_of_vectors.append(varname)

        # read the tables
        for i, n in enumerate(self.npart):

            if n != 0:
                if self.verbose and mpi.mpi_IsMaster():
                    self.message("Reading particles (type %i)..." % i)

                remaining_vectors = deepcopy(list_of_vectors)

                # loop over dataset
                block = fd["PartType%d" % i]
                if mpi.mpi_NTask() > 1:
                    init, end = self.get_particles_limits(i)

                for key in list(block.keys()):
                    varname = ntab[key]
                    if self.skip_io_block(varname):
                        self.message("Skipping %s block" % varname)
                        continue

                    if mpi.mpi_NTask() > 1:
                        data = block[key][init:end]
                    else:
                        data = block[key].value
                    if getattr(self, varname) is None:
                        setattr(self, varname, data)
                    else:
                        setattr(self, varname, np.concatenate(
                            (getattr(self, varname), data)))

                    remaining_vectors.remove(varname)

                # loop over vectors absent from the block
                for varname in remaining_vectors:
                    var = getattr(self, varname)
                    data = (
                        np.ones(
                            np.getsize(
                                n,
                                vdim[varname])) *
                        dval[varname]).astype(
                        np.float32)
                    if var is not None:
                        setattr(self, varname, np.concatenate(
                            (getattr(self, varname), data)))
                    else:
                        setattr(self, varname, data)

        dump_file = self.p_name_global[0].split("_")[:-1]
        dump_file = "".join(dump_file) + ".dump"
        data = logger.loadFromIndex(
            self.offset, dump_file, self.time_offset)

        self.pos = data["position"]
        self.vel = data["velocity"]
        self.acc = data["acceleration"]
        self.rho = data["rho"]
        self.message("Manually setting values")
        params = {
            'mh': 1.6726e-24,
            'k': None,  # useless
            'mu': 2,
            "gamma": 5. / 3.
        }
        tmp = data["entropy"]
        self.u = thermodyn.Ura(self.rho, tmp, params)
        self.num = data["id"]
        self.rsp = data["h_sph"]
        self.mass = data["mass"]

        # set tpe
        self.tpe = np.array([], np.int32)
        for i in range(len(self.npart)):
            self.tpe = np.concatenate((self.tpe, np.ones(self.npart[i]) * i))

        self.massarr = [0.] * 6



#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      BastI.py
#  brief:     
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

import numpy as np
import tarfile
import glob
import os
from pNbody.Isochrones.main import  IsochronesGrid, IsochroneBlock


def ReadOneIsochrone(f,default_keys=None):

  needtoclose=False
  
  # open if not a file descriptor  
  if type(f)==str:
    f = open(f)
    needtoclose=True
  
  # read header
  f.readline()
  line = f.readline()
  #print(line)
  
  for i in range(2):
    f.readline()
      
  header = f.readline()
  if type(header)==bytes:
    header = header.decode("utf-8")
  
  string = "[M/H] ="
  idx = header.find(string)
  MH    = float(header[idx+len(string):idx+len(string)+9])
  
  string = "Age (Myr) ="
  idx = header.find(string)
  Age   = float(header[idx+len(string):-1])
  
  #string = "Y ="
  #idx = header.find(string)
  #Y   = float(header[idx+len(string):idx+len(string)+11])
    
  f.readline()
  header = f.readline()
  if type(header)==bytes:
    header = header.decode("utf-8")
  keys = header.split()[1:]
  
  # change Mini to initial_mass
  idx = keys.index("M/Mo(ini)")
  keys[idx] = "initial_mass"

  # make a dict
  keys = {x: i for i,x in enumerate(keys)}
  
  
  # read the data
  data = np.loadtxt(f)
  
  # close file
  if needtoclose:
    f.close()

  # return
  return data, MH, Age, keys
  
  
###########################################################
# 
# Isochrone Grid
#
###########################################################

class Grid(IsochronesGrid):
  
  def Init(self):

    # get all files
    self.filenames = glob.glob(os.path.join(self.dirname,"*"))
        
    # list of Isocrhones blocks
    self.MBs = []    
        
    # read all the tar file and fill the isochrones list
    self.readAll() 
    
    # create the DB (self.M) from the isochrones list
    self.organizeDB()
    


  def readAll(self):
    '''
    Loop over all files and store the content in a list of 
    isochrones blocks.
    '''
    
    # loop over all files
    for filename in self.filenames:
      
      print(filename)

      tar = tarfile.open(filename)
      members = tar.getmembers()
      
      for member in members:
        f = tar.extractfile(member)
        
        # read   
        data, MH, Age, keys = ReadOneIsochrone(f,default_keys=None)

        # convert to a block object
        MB = self.toIsochroneBlock(data,MH,Age,keys)
        self.MBs.append(MB)


    # define keys
    self.keys = self.MBs[0].keys



  def toIsochroneBlock(self,data,MH,Age,keys):
    """
    Convert raw data from a BastI file into an isochrones block
    """
  
    MB = IsochroneBlock(data=data,keys=keys)
    
    # min/max stellar mass in solar masses
    MB.Mmin = min(MB.get("initial_mass"))
    MB.Mmax = max(MB.get("initial_mass"))
    MB.MH   =     MH
    # age in Myr
    MB.Age  =     Age
    
    return MB





#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      libmiyamoto.py
#  brief:     Defines Miyamoto profiles
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

import numpy as np
import scipy.integrate as integrate
import sys



def ComputeCircularVelocity2(V,a,Rmax,R,G=1):
  
  if R==0:
    return 0
  
  def Integrant(x):
    return -2*(V**2)/np.pi *(-x/( np.sqrt(Rmax**2-x**2)*( np.sqrt(Rmax**2-x**2) + np.sqrt(Rmax**2+a**2) )  ) - x/(x**2 + a**2))  * x/np.sqrt(R**2 -x**2)

  result = integrate.quad(Integrant, 0, R,epsrel=1e-6)  
  return result[0]



def Vcirc2(V, a, Rmax, R, G=1.):
    """
    Mestel Circular velocity
    """
    
    if Rmax is None:
      V2 = V**2 * (1- a/np.sqrt(a**2 + R**2))
    else:
      raise("we should use ComputeCircularVelocity2")
      sys.exit()
       
    return V2
    

def Vcirc(V, a, Rmax, R, G=1.):
    """
    Mestel Circular velocity
    """
    
    return np.sqrt(Vcirc2(V, a, Rmax, R, G=G))




def SurfaceDensity(V, a, Rmax, R, G=1.):
    """
    pseudo-Mestel Surface Density
    """

    return   V**2/(2*np.pi*G) /np.sqrt(R**2+ a**2)




def Omega2(V, a, Rmax, R, G=1.):
    """
    Omega2
    """
    Om2 = Vcirc2(V, a, Rmax, R, G=G) / R**2
    return Om2


def Omega(V, a, Rmax, R, G=1.):
    """
    Omega
    """
    return np.sqrt(Omega2(V, a, Rmax, R, G=G))



def Kappa2(V, a, Rmax, R, G=1.):
    """
    Kappa2
    """
    K2 =  V**2 * a /(R**2+a**2)**(3/2.)    +   2*Vcirc2(V, a, Rmax, R, G=G) / R**2
    return K2


def Kappa(V, a, Rmax, R, G=1.):
    """
    Kappa
    """
    return np.sqrt(Kappa2(V, a, Rmax, R, G=G))








#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      libqt.py
#  brief:     PyQT4 interface
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

import sys

from PyQt5 import QtCore, QtGui
from PyQt5.QtWidgets import QApplication, QMainWindow, QLabel, QGridLayout, QWidget
from PyQt5.QtGui import QPixmap, QImage


import numpy as np
from .libutil import *

from PIL import ImageQt


def rgb(r, g, b):
    # use qRgb to pack the colors, and then turn the resulting long
    # into a negative integer with the same bit pattern.
    return (QtGui.qRgb(r, g, b) & 0xffffff) - 0x1000000


class QNumarrayImage(QtGui.QImage):
    """
    QNumarrayImage class
    """

    def __init__(self, data, palette_name):

        format = QtGui.QImage.Format_Indexed8

        # include the palette
        palette = Palette(palette_name)

        # data = data*0+64
        data = data.astype(np.uint8)

        # give the right form
        shape = data.shape
        data = np.transpose(data)
        data = np.ravel(data)
        data.shape = shape

        self.__data = data.tostring()

        # init object
        QtGui.QImage.__init__(
            self,
            self.__data,
            data.shape[0],
            data.shape[1],
            format)

        # color table
        colortable = []
        for i in range(256):
            colortable.append(QtGui.qRgb(palette.r[i], palette.g[i], palette.b[i]))
        self.setColorTable(colortable)


def qtplot(mat, palette='light'):
    """
    plot a matrix using qt
    """

    app = QtGui.QApplication(sys.argv)

    imageLabel = QtGui.QLabel()
    imageLabel.setScaledContents(True)

    matint, mn, mx, cd = set_ranges(
        mat, scale='lin', mn=None, mx=None, cd=None)

    # without qt
    imageQt = QNumarrayImage(matint, palette)

    # using PIL
    # imagePIL = get_image(matint,name=None,palette_name=palette)
    # imageQt  = ImageQt.ImageQt(imagePIL)

    imageLabel.setPixmap(QtGui.QPixmap.fromImage(imageQt))

    imageLabel.show()
    sys.exit(app.exec_())




class DisplayWindow(QWidget):

    def __init__(self,imagePIL):
        super().__init__()
                
        # this no longer works !!!
        #imageQt = ImageQt.ImageQt(imagePIL)   
        #self.im = QPixmap.fromImage(imageQt)
        
        # so, do things in detail through numpy
        imagePIL = imagePIL.convert("RGB")
        data = np.array(imagePIL)
        qimage = QImage(data.data, data.shape[1], data.shape[0], data.strides[0], QImage.Format_RGB888);
        self.im = QPixmap(qimage)


        self.label = QLabel()
        self.label.setPixmap(self.im)

        self.grid = QGridLayout()
        self.grid.addWidget(self.label,1,1)
        self.setLayout(self.grid)

        self.setGeometry(50,50,320,200)
        self.setWindowTitle("PyQT show image")
        self.show()


def display(imagePIL):

  app = QApplication(sys.argv)
  ex = DisplayWindow(imagePIL)
  app.exec_()
  #sys.exit(app.exec_())














###########################################################################################
#  package:   Mockimgs
#  file:      telescope.py
#  brief:     telescope class
#  copyright: GPLv3
#             Copyright (C) 2023 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

import astropy.units
from astropy import units as u

class Telescope():
  '''
  Define the telescope class.
  '''
  
  def __init__(self,name=None,focal=None): 
       
    self.name  = name
    self.set_focal(focal)
    
  def info(self):
    """
    give info on the telescope
    """  
    print("telescope name     : %s"%self.name)


  def get_parameters(self,fmt=None):
    """
    return a dictionary containing usefull parameters    
    """
    params = {}
    if fmt=="fits":
      params["TELNAME"]            = (self.name,"Telescope name")
    else:
      params["telescope_name"]     = self.name

    if fmt=="fits":
      if self.focal is None:
        params["FOCAL"]              = str(None)
      else:  
        params["FOCAL"]              = (self.focal.to(u.cm).value,"Telescope focal in cm")
    else:
      params["telescope_focal"]    = self.focal
    
    return params


      
  def get_focal(self):
    """
    return the telescope focal
    """    
    return self.focal
      

  def set_focal(self,focal):
    """
    set the telescope focal
    
    focal : an astropy quantity in unit length
    """   
    if focal is not None:
      if type(focal) != astropy.units.quantity.Quantity:
        raise TypeError("focal must be of type astropy.units.quantity.Quantity") 
    
    self.focal = focal
    
    
    
    
    



#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      ctes.py
#  brief:     Physics Constants
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################


from . import units as u


GRAVITY = u.PhysCte(6.6732e-11, u.Unit_G)
SOLAR_MASS = u.PhysCte(1.989e33, u.Unit_g)
SOLAR_LUM = u.PhysCte(3.826e33, u.Unit_erg / u.Unit_s)
AVOGADRO = u.PhysCte(6.0220e+23, u.Unit_mol)
BOLTZMANN = u.PhysCte(1.3807e-23, u.Unit_J / u.Unit_K)
GAS_CONST = u.PhysCte(8.3144e+00, u.Unit_J / u.Unit_mol)
C = u.PhysCte(2.99792458e8, u.Unit_m / u.Unit_s)
PLANCK = u.PhysCte(6.6262e-34, u.Unit_J * u.Unit_s)
PROTONMASS = u.PhysCte(1.6726e-27, u.Unit_kg)
ELECTRONMASS = u.PhysCte(9.1095e-31, u.Unit_kg)
ELECTRONCHARGE = u.PhysCte(4.8032e-10, u.Unit_C)
AV = u.PhysCte(6.828e-50, u.Unit_Pa * u.Unit_m**6)
BV = u.PhysCte(4.419e-29, u.Unit_m**3)
HUBBLE = u.PhysCte(3.2407789e-18, 1 / u.Unit_s)  # in h/s



def convert_ctes(units):
    """
    convert a constant into a given unit system.
    """
    global BOLTZMANN, PROTONMASS

    # UnitLength_in_cm = units[0]
    UnitMass_in_g = units[1]
    UnitVelocity_in_cm_per_s = units[2]
    # UnitTime_in_s = UnitLength_in_cm / UnitVelocity_in_cm_per_s
    UnitEnergy_in_cgs = UnitMass_in_g * UnitVelocity_in_cm_per_s**2

    BOLTZMANN = BOLTZMANN / UnitEnergy_in_cgs
    PROTONMASS = PROTONMASS / UnitMass_in_g


###########################################################################################
#  package:   Mockimgs
#  file:      utils.py
#  brief:     some useful routines
#  copyright: GPLv3
#             Copyright (C) 2023 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

import numpy as np
from scipy.spatial.transform import Rotation



def convert_rsp_options_to_parameters(opt):
  '''
  convert argparse rsp options to parameters
  stored in a dictionary 
  '''
  rsp_opts = {}
  rsp_opts["mode"] = opt.rsp_mode
  rsp_opts["val"]  = opt.rsp_val
  rsp_opts["max"]  = opt.rsp_max
  rsp_opts["sca"]  = opt.rsp_sca  
  rsp_opts["fac"]  = opt.rsp_fac  
  return rsp_opts
  


def align(pos,axis,axis_ref=[0,0,1]):
  """
  rotate the model with a rotation that align axis with axis_ref
  
  pos      : a set of positions
  axis     : the axis
  axis_ref : the reference axis
  """
  axis1 = axis_ref
  axis2 = axis

  a1 = np.array(axis1, float)
  a2 = np.array(axis2, float)

  a3 = np.array([0, 0, 0], float)
  a3[0] = a1[1] * a2[2] - a1[2] * a2[1]
  a3[1] = a1[2] * a2[0] - a1[0] * a2[2]
  a3[2] = a1[0] * a2[1] - a1[1] * a2[0]

  n1 = np.sqrt(a1[0]**2 + a1[1]**2 + a1[2]**2)
  n2 = np.sqrt(a2[0]**2 + a2[1]**2 + a2[2]**2)
  angle = -np.arccos(np.inner(a1, a2) / (n1 * n2))

  if angle==0:
    return pos
    
  # set the norm of a3 to be angle
  norm = np.sqrt(a3[0]**2 + a3[1]**2 + a3[2]**2)
  a3 = a3/norm*angle  
    
  # create the rotation matix
  R = Rotation.from_rotvec(a3)
  
  # do the rotation
  pos = R.apply(pos)
  
  return pos

  



###########################################################################################
#  package:   pNbody
#  file:      cut.py
#  brief:
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

from Nbody import *
from Numeric import *

xm = self.params.get('size')[0]
zm = self.params.get('size')[1]
xmin = - xm
xmax = + xm
zmin = - zm
zmax = + zm

# expose the model
self.nbodyviewer.pose()

# keep only particules that are in the window
self.X = self.X.selectp(self.nbodyviewer.X.num)


x = self.nbodyviewer.X.pos[:, 0]
z = self.nbodyviewer.X.pos[:, 2]

cx1 = greater(x, xmin)
cx2 = less(x, xmax)

cz1 = greater(z, zmin)
cz2 = less(z, zmax)

self.X = self.X.selectc(cx1 * cx2 * cz1 * cz2)

# redisplay
self.display()
self.display_info(self.X)



###########################################################################################
#  package:   pNbody
#  file:      gadget.py
#  brief:     Gadget format (binary)
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

##########################################################################
#
# GADGET CLASS
#
##########################################################################

import sys
import os
import string
import types
from pNbody import mpiwrapper as mpi
from pNbody import errorfuncs as error
from pNbody import iofunc as io 
from pNbody import libutil
from pNbody import parameters
from pNbody import units, ctes, cosmo, thermodyn

import numpy as np
# define some units

try:				# all this is usefull to read files
    from mpi4py import MPI
except BaseException:
    MPI = None


class Nbody_gadget:

    def _init_spec(self):
        # define some units
        Unit_atom = ctes.PROTONMASS.into(units.cgs) * units.Unit_g
        Unit_atom.set_symbol('atom')
        out_units = units.UnitSystem(
            'local', [units.Unit_cm, Unit_atom, units.Unit_s, units.Unit_K])
        self.DensityUnit_in_AtomPerCC = out_units.UnitDensity

    def get_excluded_extension(self):
        """
        Return a list of file to avoid when extending the default class.
        """
        return []
        
    def getParticleMatchingDict(self):
        """
        Return a list of file to avoid when extending the default class.
        """
        
        index = {
            'gas':    0,
            'halo':   2,
            'disk':   2,
            'bulge':  3,
            'stars':  1,
            'bndry':  5,
            'stars1': 1,
            'halo1':  2}
        
        return index      
        

    def check_spec_ftype(self):
        # read the first 4 bites and check that they are equal to 256
        for name in self.p_name:
            mpi_test = mpi.mpi_NTask() > 1 and self.pio == "no"
            if mpi.mpi_IsMaster() or self.pio == "yes":
                f = open(name, 'rb')

                try:
                    htype = np.int32
                    nb1 = np.frombuffer(f.read(4), htype)
                    if sys.byteorder != sys.byteorder:
                        nb1.byteswap(True)
                    nb1 = nb1[0]
                except BaseException:
                    if mpi_test:
                        mpi.mpi_bcast(False)
                    f.close()
                    raise error.FormatError("gadget")

                if nb1 != 256:
                    if mpi_test:
                        mpi.mpi_bcast(False)
                    f.close()
                    raise error.FormatError("gadget")
                elif mpi_test:
                    mpi.mpi_bcast(True)
                f.close()
            else:
                test = mpi.mpi_bcast(None)
                if test is False:
                    raise error.FormatError("gadget")

    def get_read_fcts(self):
        return [self.read_particles]

    def get_write_fcts(self):
        return [self.write_particles]

    def get_mxntpe(self):
        return 6

    def get_default_spec_vars(self):
        '''
        return specific variables default values for the class
        '''
        return {'massarr': np.array([0, 0, self.nbody, 0, 0, 0]),
                'atime': 0.,
                'redshift': 0.,
                'flag_sfr': 0,
                'flag_feedback': 0,
                'nall': np.array([0, 0, self.nbody, 0, 0, 0]),
                'flag_cooling': 0,
                'num_files': 1,
                'boxsize': 0.,
                'omega0': 0.,
                'omegalambda': 0.,
                'hubbleparam': 0.,
                'flag_age': 0.,
                'hubbleparam': 0.,
                'flag_metals': 0.,
                'nallhw': np.array([0, 0, 0, 0, 0, 0]),
                'flag_entr_ic': 0,
                'flag_chimie_extraheader': 0,
                'flag_thermaltime': 0,
                'critical_energy_spec': 0.,
                'empty': '',
                'comovingintegration': None,
                'hubblefactorcorrection': None,
                'comovingtoproperconversion': False,
                'utype':"gear",
                }

    def initComovingIntegration(self):
        """
        set true if the file has been runned using
        the comoving integration scheme
        
        this function is specific to the gadget format
        """

        # do nothing, if the value is already set
        if self.comovingintegration is not None:
            return

        flag = True

        # this is not very good, however, there is not other choice...
        if self.omega0 == 0 and self.omegalambda == 0:
            flag = False

        self.comovingintegration = flag
        
    def initHubbleFactorCorrection(self):
        """
        set true if physical quantities must be corrected from the Hubble parameter
        
        this function is specific to the gadget format
        """

        if self.comovingintegration==False or self.hubbleparam==1:
          self.hubblefactorcorrection = False
        else:  
          self.hubblefactorcorrection = True  


    def get_massarr_and_nzero(self):
        """
        return massarr and nzero

        !!! when used in //, if only a proc has a star particle,
        !!! nzero is set to 1 for all cpu, while massarr has a length of zero !!!
        """

        if self.has_var('massarr') and self.has_var('nzero'):
            if self.massarr is not None and self.nzero is not None:
                self.warning("warning : get_massarr_and_nzero : here we use massarr and nzero %s %s"%(self.massarr,self.nzero))
                return self.massarr, self.nzero

        massarr = np.zeros(len(self.npart), float)
        nzero = 0

        # for each particle type, see if masses are equal
        for i in range(len(self.npart)):
            first_elt = sum((np.arange(len(self.npart)) < i) * self.npart)
            last_elt = first_elt + self.npart[i]

            if first_elt != last_elt:
                c = (self.mass[first_elt] ==
                     self.mass[first_elt:last_elt]).astype(int)
                if sum(c) == len(c):
                    massarr[i] = self.mass[first_elt]
                else:
                    nzero = nzero + len(c)

        return massarr.tolist(), nzero


    def read_particles(self, f):
        '''
        read gadget file
        '''
        ##########################################
        # read the header and send it to each proc
        ##########################################
        tpl = (
            24,
            48,
            float,
            float,
            np.int32,
            np.int32,
            24,
            np.int32,
            np.int32,
            float,
            float,
            float,
            float,
            np.int32,
            np.int32,
            24,
            np.int32,
            np.int32,
            np.int32,
            float,
            44)
        header = io.ReadBlock(f, tpl, byteorder=self.byteorder, pio=self.pio)
        npart, massarr, atime, redshift, flag_sfr, flag_feedback, nall, flag_cooling, num_files, boxsize, omega0, omegalambda, hubbleparam, flag_age, flag_metals, nallhw, flag_entr_ic, flag_chimie_extraheader, flag_thermaltime, critical_energy_spec, empty = header

        if np.fabs(
                flag_chimie_extraheader) > 1:		# if the header is empty, it may be indetermined
            flag_chimie_extraheader = 0

        npart = np.frombuffer(npart, np.int32)
        massarr = np.frombuffer(massarr, float)
        nall = np.frombuffer(nall, np.int32)
        nallhw = np.frombuffer(nallhw, np.int32)

        if sys.byteorder != self.byteorder:
            npart.byteswap(True)
            massarr.byteswap(True)
            nall.byteswap(True)
            nallhw.byteswap(True)

        if flag_metals:		# gas metal properties
            NELEMENTS = flag_metals
            self.NELEMENTS = NELEMENTS
        else:
            NELEMENTS = 0
            self.NELEMENTS = NELEMENTS

        if flag_thermaltime != 1:
            flag_thermaltime = 0

        ##########################################
        # computes nzero
        ##########################################
        # count number of particles that have non constant masses and then
        # have masses storded further

        if self.pio == 'no':

            npart_tot = npart
            npart_all = libutil.get_npart_all(npart, mpi.mpi_NTask())
            npart = npart_all[mpi.mpi_ThisTask()]			# local
            npart_read = npart_tot
            nbody_read = sum(npart_read)

            npart_m_read = npart_tot * (massarr == 0)

            ngas = npart[0]
            ngas_read = npart_tot[0]
            nstars = npart[1]
            nstars_read = npart_tot[1]

            # compute nzero
            nzero = 0
            mass = np.array([])
            for i in range(len(npart_tot)):
                if massarr[i] == 0:
                    nzero = nzero + npart_tot[i]
                else:
                    mass = np.concatenate(
                        (mass, np.ones(npart[i]) * massarr[i]))

        else:

            npart_tot = mpi.mpi_allreduce(npart)
            npart_all = None						# each proc read for himself
            npart = npart						# local
            npart_read = None						# each proc read for himself
            nbody_read = sum(npart)

            npart_m_read = None						# each proc read for himself

            ngas = npart[0]
            ngas_read = ngas
            nstars = npart[1]
            nstars_read = nstars

            # compute nzero
            nzero = 0
            mass = np.array([])
            for i in range(len(npart)):
                if massarr[i] == 0:
                    nzero = nzero + npart[i]
                else:
                    mass = np.concatenate(
                        (mass, np.ones(npart[i]) * massarr[i]))

        nbody = sum(npart)
        nbody_tot = sum(npart_tot)

        if npart_m_read is not None:
            if sum(npart_m_read) != nzero:
                raise Exception(
                    "sum(npart_m) (%d) != nzero (%d)" %
                    (sum(npart_m_read), nzero)(npart_m_read))

        ##########################################
        # optionnally read extra header
        ##########################################
        if flag_chimie_extraheader == 1:
            self.message("reading chimie extra-header...")
            tpl = (np.int32, int(NELEMENTS) * 4, 256 - 4 - int(NELEMENTS) * 4)

            nelts, ChimieSolarMassAbundances, labels = io.ReadBlock(
                f, tpl, byteorder=self.byteorder, pio=self.pio)

            labels = labels.decode("utf-8")

            nelts = int(nelts)
            self.ChimieNelements = nelts

            ChimieElements = str.split(labels, ',')[:nelts]
            self.ChimieElements = ChimieElements

            ChimieSolarMassAbundances = np.frombuffer(
                ChimieSolarMassAbundances, np.float32)
            self.ChimieSolarMassAbundances = {}
            for i, elt in enumerate(self.ChimieElements):
                self.ChimieSolarMassAbundances[elt] = ChimieSolarMassAbundances[i]
        else:
            self.ChimieNelements = int(NELEMENTS)
            self.ChimieElements = ['Fe', 'Mg', 'O', 'Metals']

            self.ChimieSolarMassAbundances = {}
            self.ChimieSolarMassAbundances['Fe'] = 0.001771
            self.ChimieSolarMassAbundances['Mg'] = 0.00091245
            self.ChimieSolarMassAbundances['O'] = 0.0108169
            self.ChimieSolarMassAbundances['Metals'] = 0.02

        ##########################################
        # read and send particles attribute
        ##########################################
        vec = 'pos'
        self.message("reading %s..." % vec)
        pos = io.ReadDataBlock(f, np.float32, shape=(nbody_read, 3),
                               byteorder=self.byteorder, pio=self.pio,
                               npart=npart_read,
                               skip=self.skip_io_block(vec))

        vec = 'vel'
        self.message("reading %s..." % vec)
        vel = io.ReadDataBlock(f, np.float32, shape=(nbody_read, 3),
                               byteorder=self.byteorder, pio=self.pio,
                               npart=npart_read,
                               skip=self.skip_io_block(vec))

        vec = 'num'
        self.message("reading %s..." % vec)
        num = io.ReadDataBlock(f, np.int32, shape=(nbody_read,),
                               byteorder=self.byteorder, pio=self.pio,
                               npart=npart_read,
                               skip=self.skip_io_block(vec))

        ##########################################
        # read mass if needed
        ##########################################
        if nzero != 0:

            vec = 'mass'
            self.message("reading %s..." % vec)
            massnzero = io.ReadDataBlock(
                f, np.float32, shape=(nzero,),
                byteorder=self.byteorder, pio=self.pio,
                npart=npart_m_read, skip=self.skip_io_block(vec))

            if nzero == nbody_tot:
                mass = massnzero
            else:
                mass = np.array([])
                i1 = 0
                i2 = None
                for i in range(len(npart)):
                    if npart[i] != 0:		# particles belong to the class
                        if massarr[i] != 0:
                            mass = np.concatenate(
                                (mass, np.ones(npart[i]) * massarr[i]))
                        else:
                            i2 = i1 + npart[i]
                            mass = np.concatenate((mass, massnzero[i1:i2]))
                            i1 = i2

                if i2 is not None and i2 != len(massnzero):
                    raise Exception("i2=", i2, """!=len(massnzero)""")

                if len(mass) != nbody:				# if pio, we should maybee put nbody_tot ???
                    raise Exception("len(mass)=", len(mass), "!=nbody")

        '''
	  if massarr[i] == 0:
	    if npart[i]!=0:
	      if len(massnzero)!=npart[i]:
		raise "this case is not taken into account, sorry !"
      	      mass = concatenate((mass,massnzero))
	  else:
      	    mass = concatenate((mass,ones(npart[i])*massarr[i]))
    '''

        # extentions
        u = None
        rho = None
        rsp = None
        opt1 = None
        opt2 = None
        erd = None
        dte = None
        pot = None
        acc = None
        tstar = None
        minit = None
        idp = None
        metals = None
        thtsnii = None
        thtsnia = None

        if not io.end_of_file(f, pio=self.pio, MPI=MPI) and ngas_read != 0:
            vec = 'u'
            self.message("reading %s..." % vec)
            u = io.ReadDataBlock(
                f, np.float32, shape=(ngas_read,),
                byteorder=self.byteorder, pio=self.pio,
                npart=None, skip=self.skip_io_block(vec))
            
            if u is not None:
                u = np.concatenate(
                    (u, np.zeros(nbody - ngas).astype(np.float32))
                )

        if not io.end_of_file(f, pio=self.pio, MPI=MPI) and ngas_read != 0:
            vec = 'rho'
            self.message("reading %s..." % vec)
            rho = io.ReadDataBlock(
                f, np.float32, shape=(ngas_read,),
                byteorder=self.byteorder, pio=self.pio,
                npart=None, skip=self.skip_io_block(vec))
            if rho is not None:
                rho = np.concatenate(
                    (rho, np.zeros(nbody - ngas).astype(np.float32)))

        if not io.end_of_file(f, pio=self.pio, MPI=MPI) and ngas_read != 0:
            vec = 'rsp'
            self.message("reading %s..." % vec)
            rsp = io.ReadDataBlock(
                f, np.float32, shape=(ngas_read,),
                byteorder=self.byteorder, pio=self.pio,
                npart=None, skip=self.skip_io_block(vec))

            if rsp is not None:
                rsp = np.concatenate(
                    (rsp, np.zeros(nbody - ngas).astype(np.float32))
                )

        # here it is the end of the minimal output

        if flag_metals:		# gas metal properties

            if not io.end_of_file(f, pio=self.pio, MPI=MPI):
                vec = 'metals'
                self.message("reading %s..." % vec)
                metals = io.ReadDataBlock(
                    f, np.float32, shape=(ngas_read, NELEMENTS),
                    byteorder=self.byteorder, pio=self.pio,
                    npart=None, skip=self.skip_io_block(vec))
                if metals is not None:
                    metals = np.concatenate(
                        (metals, np.zeros((nbody - ngas, NELEMENTS)).astype(np.float32))
                    )

        if flag_thermaltime:  # gas thermal(adiabatic) time

            if not io.end_of_file(f, pio=self.pio, MPI=MPI) and ngas_read != 0:
                vec = 'thtsnii'
                self.message("reading %s..." % vec)
                thtsnii = io.ReadDataBlock(
                    f, np.float32, shape=(ngas_read,),
                    byteorder=self.byteorder, pio=self.pio,
                    npart=None, skip=self.skip_io_block(vec))

                if thtsnii is not None:
                    thtsnii = np.concatenate(
                        (thtsnii, np.zeros(nbody - ngas).astype(np.float32)))

            if not io.end_of_file(f, pio=self.pio, MPI=MPI) and ngas_read != 0:
                vec = 'thtsnia'
                self.message("reading %s..." % vec)
                thtsnia = io.ReadDataBlock(
                    f, np.float32, shape=(ngas_read,),
                    byteorder=self.byteorder, pio=self.pio,
                    npart=None, skip=self.skip_io_block(vec))

                if thtsnia is not None:
                    thtsnia = np.concatenate(
                        (thtsnia, np.zeros(nbody - ngas).astype(np.float32)))

        if flag_age:		# stellar properties

            if not io.end_of_file(f, pio=self.pio,MPI=MPI) and nstars_read != 0:
                vec = 'tstar'
                self.message("reading %s..." % vec)
                tstar = io.ReadDataBlock(
                    f, np.float32, shape=(nstars_read,),
                    byteorder=self.byteorder, pio=self.pio,
                    npart=None, skip=self.skip_io_block(vec))

                if tstar is not None:
                    tstar = np.concatenate(
                        (-1 * np.ones(ngas).astype(np.float32),
                         tstar, -1 * np.ones(nbody - ngas - nstars).astype(np.float32))
                    )

            if (not io.end_of_file(f, pio=self.pio, MPI=MPI)
                    and nstars_read != 0):
                vec = 'minit'
                self.message("reading %s..." % vec)
                minit = io.ReadDataBlock(
                    f, np.float32, shape=(nstars_read,),
                    byteorder=self.byteorder, pio=self.pio,
                    npart=None, skip=self.skip_io_block(vec))

                if minit is not None:
                    minit = np.concatenate(
                        (np.zeros(ngas).astype(np.float32),
                         minit,
                         np.zeros(nbody - ngas - nstars).astype(np.float32))
                    )

            if not io.end_of_file(f, pio=self.pio, MPI=MPI) and nstars_read != 0:
                vec = 'idp'
                self.message("reading %s..." % vec)
                idp = io.ReadDataBlock(
                    f, np.int32, shape=(nstars_read,),
                    byteorder=self.byteorder, pio=self.pio,
                    npart=None, skip=self.skip_io_block(vec))

                if idp is not None:
                    idp = np.concatenate((-1 * np.ones(ngas).astype(np.float32),
                                          idp, -1 * np.ones(nbody - ngas - nstars).astype(np.float32)))

            if not io.end_of_file(f, pio=self.pio, MPI=MPI) and nstars_read != 0:
                vec = 'rho_stars'
                self.message("reading %s..." % vec)
                rho_stars = io.ReadDataBlock(
                    f, np.float32, shape=(nstars_read,),
                    byteorder=self.byteorder, pio=self.pio,
                    npart=None, skip=self.skip_io_block('rho'))
                if rho_stars is not None:
                    if rho is None:
                        rho = np.concatenate(
                            (rho_stars,
                             np.zeros(nbody - ngas - nstars).astype(np.float32)))
                    else:
                        rho = np.concatenate((rho[:ngas], rho_stars, np.zeros(
                            nbody - ngas - nstars).astype(np.float32)))

            if not io.end_of_file(f, pio=self.pio, MPI=MPI) and nstars_read != 0:
                vec = 'rsp_stars'
                self.message("reading %s..." % vec)
                rsp_stars = io.ReadDataBlock(
                    f, np.float32, shape=(nstars_read,),
                    byteorder=self.byteorder, pio=self.pio,
                    npart=None, skip=self.skip_io_block('rsp'))
                
                if rsp_stars is not None:
                    if rsp is None:
                        rsp = np.concatenate(
                            (rsp_stars,
                             np.zeros(nbody - ngas - nstars).astype(np.float32))
                        )
                    else:
                        rsp = np.concatenate((rsp[:ngas], rsp_stars, np.zeros(
                            nbody - ngas - nstars).astype(np.float32)))

        if flag_metals:	      # stars metal properties

            if not io.end_of_file(f, pio=self.pio, MPI=MPI) and nstars_read != 0:
                vec = 'metals_stars'
                self.message("reading %s..." % vec)
                metals_stars = io.ReadDataBlock(
                    f, np.float32, shape=(nstars_read, NELEMENTS),
                    byteorder=self.byteorder, pio=self.pio,
                    npart=None, skip=self.skip_io_block('metals'))

                if metals_stars is not None:
                    metals = np.concatenate((metals[:ngas, :], metals_stars, np.zeros(
                        (nbody - ngas - nstars, NELEMENTS)).astype(np.float32)))

        # other variables
        if not io.end_of_file(f, pio=self.pio, MPI=MPI):
            pass

        if not io.end_of_file(f, pio=self.pio, MPI=MPI):
            vec = 'opt1'
            self.message("reading %s..." % vec)
            opt1 = io.ReadDataBlock(
                f,
                np.float32,
                shape=(
                    ngas_read,
                ),
                byteorder=self.byteorder,
                pio=self.pio,
                npart=None,
                skip=self.skip_io_block(vec))
            if opt1 is not None:
                if (len(opt1) == ngas):
                    opt1 = np.concatenate(
                        (opt1,
                         np.zeros(
                             nbody -
                             ngas).astype(
                             np.float32)))

        if not io.end_of_file(f, pio=self.pio, MPI=MPI):
            vec = 'opt2'
            self.message("reading %s..." % vec)
            opt2 = io.ReadDataBlock(
                f,
                np.float32,
                shape=(
                    ngas_read,
                ),
                byteorder=self.byteorder,
                pio=self.pio,
                npart=None,
                skip=self.skip_io_block(vec))
            if opt2 is not None:
                if (len(opt2) == ngas):
                    opt2 = np.concatenate(
                        (opt2,
                         np.zeros(
                             nbody -
                             ngas).astype(
                             np.float32)))

        if not io.end_of_file(f, pio=self.pio, MPI=MPI):
            vec = 'erd'
            self.message("reading %s..." % vec)
            erd = io.ReadDataBlock(
                f,
                np.float32,
                shape=(
                    ngas_read,
                ),
                byteorder=self.byteorder,
                pio=self.pio,
                npart=None,
                skip=self.skip_io_block(vec))
            if (len(erd) == ngas):
                erd = np.concatenate(
                    (erd,
                     np.zeros(
                         nbody -
                         ngas).astype(
                         np.float32)))

        if not io.end_of_file(f, pio=self.pio, MPI=MPI):
            vec = 'dte'
            self.message("reading %s..." % vec)
            dte = io.ReadDataBlock(
                f,
                np.float32,
                shape=(
                    ngas_read,
                ),
                byteorder=self.byteorder,
                pio=self.pio,
                npart=None,
                skip=self.skip_io_block(vec))
            dte = np.concatenate(
                (dte,
                 np.zeros(
                     nbody -
                     ngas).astype(
                     np.float32)))

        # if not io.end_of_file(f,pio=self.pio,MPI=MPI):
        #  pot  = io.ReadDataBlock(f,np.float32,shape=(ngas_read,),byteorder=self.byteorder,pio=self.pio,npart=None)

        # make global
        self.npart = npart
        self.massarr = massarr

        self.atime = atime
        self.redshift = redshift
        self.flag_sfr = flag_sfr
        self.flag_feedback = flag_feedback
        self.nall = nall
        self.flag_cooling = flag_cooling
        self.num_files = num_files
        self.boxsize = boxsize
        self.omega0 = omega0
        self.omegalambda = omegalambda
        self.hubbleparam = hubbleparam
        self.flag_age = flag_age
        self.flag_metals = flag_metals
        self.flag_thermaltime = flag_thermaltime
        self.nallhw = nallhw
        self.flag_entr_ic = flag_entr_ic
        self.flag_chimie_extraheader = flag_chimie_extraheader
        self.critical_energy_spec = critical_energy_spec
        self.empty = empty
        self.nbody = nbody

        self.pos = pos
        self.vel = vel
        self.mass = mass
        self.num = num

        self.tpe = np.array([], np.int32)
        for i in range(len(npart)):
            self.tpe = np.concatenate((self.tpe, np.ones(npart[i]) * i))

        self.nzero = nzero
        self.u = u
        self.rho = rho

        self.tstar = tstar
        self.minit = minit
        self.idp = idp
        self.metals = metals

        self.thtsnii = thtsnii
        self.thtsnia = thtsnia

        self.opt1 = opt1
        self.opt2 = opt2
        self.erd = erd
        self.dte = dte
        self.pot = pot
        self.acc = acc

        self.rsp = rsp

        if isinstance(self.massarr, np.ndarray):
            self.massarr = self.massarr.tolist()
        if isinstance(self.nall, np.ndarray):
            self.nall = self.nall.tolist()
        if isinstance(self.nallhw, np.ndarray):
            self.nallhw = self.nallhw.tolist()

        # init comoving integration
        self.comovingintegration = None
        self.initComovingIntegration()
        self.initHubbleFactorCorrection()
        
        # set 
        if self.comovingintegration:
          self.comovingtoproperconversion=True
        else:
          self.comovingtoproperconversion=False  
        

    def write_particles(self, f):
        '''
        specific format for particle file
        '''

        # here, we must let the user decide if we creates
        # the mass block or not, event if all particles have the same mass
        massarr, nzero = self.get_massarr_and_nzero()

        if self.pio == 'yes':

            '''
            here, we have to compute also
            mass for each proc
            '''

            npart = self.npart
            nall = self.npart_tot
            num_files = mpi.NTask

            npart_write = None
            npart_m_write = None

        else:

            npart = self.npart
            nall = self.npart_tot
            npart_all = np.array(mpi.mpi_allgather(npart))
            num_files = 1

            npart_write = self.npart
            npart_m_write = np.array(self.npart) * \
                (np.array(self.massarr) == 0)

            # compute the global massarr and global nzero
            nzero_tot = mpi.mpi_sum(nzero)
            massarr_all = np.array(mpi.mpi_allgather(massarr))
            massarr_tot = np.zeros(len(npart), float)

            for i in range(len(npart)):

                # keep only values where there are particles
                massarr_all_red = np.compress(
                    npart_all[:, i] != 0, massarr_all[:, i])

                if len(massarr_all_red) > 0:
                    if (massarr_all_red == massarr_all_red).all():
                        massarr_tot[i] = massarr_all[0, i]
                    else: 		  # not equal
                        raise Exception("this case is not implemented")
                        massarr_tot[i] = 0.0
                        nzero_tot = nzero_tot + sum(npart_write[:, i])

            # now, re-compute nzero for the current node
            massarr = massarr_tot
            nzero = 0
            for i in range(len(npart)):
                if massarr[i] == 0:
                    nzero = nzero + npart[i]

        # now that we have the right massarr and nzero,
        # we can compute massnzero for each node
        nzero_all = np.zeros((mpi.NTask, len(self.npart)))

        if nzero != 0:
            ni = 0
            massnzero = np.array([], np.float32)
            for i in range(len(self.npart)):
                if npart[i] != 0 and massarr[i] == 0.:
                    massnzero = np.concatenate(
                        (massnzero, self.mass[ni:ni + self.npart[i]]))
                    nzero_all[mpi.ThisTask,
                              i] = nzero_all[mpi.ThisTask,
                                             i] + npart[i]

                ni = ni + self.npart[i]

        nzero_all = mpi.mpi_allreduce(nzero_all)

        if self.pio == 'yes':
            if nzero != 0 and len(
                    massnzero) == 0:		# !!! because zere is a bug see warning in get_massarr_and_nzero
                nzero = 0
        else:
            npart = self.npart_tot
            nzero = mpi.mpi_allreduce(nzero)  # to ensure that all nodes
            # will do -> write mass if needed

        # header
        if sys.byteorder == self.byteorder:
            npart = np.array(npart, np.int32).tostring()
            massarr = np.array(massarr, float).tostring()
        else:
            npart = np.array(npart, np.int32).byteswap().tostring()
            massarr = np.array(massarr, float).byteswap().tostring()

        atime = self.atime
        redshift = self.redshift
        flag_sfr = self.flag_sfr
        flag_feedback = self.flag_feedback

        if sys.byteorder == self.byteorder:
            nall = np.array(nall, np.int32).tostring()
        else:
            nall = np.array(nall, np.int32).byteswap().tostring()

        flag_cooling = self.flag_cooling
        num_files = num_files
        boxsize = self.boxsize
        omega0 = self.omega0
        omegalambda = self.omegalambda
        hubbleparam = self.hubbleparam
        flag_age = self.flag_age
        flag_metals = self.flag_metals

        if sys.byteorder == self.byteorder:
            nallhw = np.array(self.nallhw, float).tostring()
        else:
            nallhw = np.array(self.nallhw, float).byteswap().tostring()

        flag_entr_ic = self.flag_entr_ic
        flag_chimie_extraheader = self.flag_chimie_extraheader
        flag_thermaltime = self.flag_thermaltime
        critical_energy_spec = self.critical_energy_spec
        empty = self.empty

        # header
        tpl = (
            (npart, 24),
            (massarr, 48),
            (atime, float),
            (redshift, float),
            (flag_sfr, np.int32),
            (flag_feedback, np.int32),
            (nall, 24),
            (flag_cooling, np.int32),
            (num_files, np.int32),
            (boxsize, float),
            (omega0, float),
            (omegalambda, float),
            (hubbleparam, float),
            (flag_age, np.int32),
            (flag_metals, np.int32),
            (nallhw, 24),
            (flag_entr_ic, np.int32),
            (flag_chimie_extraheader, np.int32),
            (flag_thermaltime, np.int32),
            (critical_energy_spec, float),
            (empty, 44))
        io.WriteBlock(f, tpl, byteorder=self.byteorder)

        # extra header
        if self.flag_chimie_extraheader:
            self.message("writing chimie extra-header...")

            SolarMassAbundances = np.zeros(self.ChimieNelements, np.float32)
            labels = ""
            for i, elt in enumerate(self.ChimieElements):
                SolarMassAbundances[i] = self.ChimieSolarMassAbundances[elt]
                labels = labels + "%s," % elt

            labels_len = (256 - 4 - self.ChimieNelements * 4)
            labels = labels + (labels_len - len(labels)) * " "
            labels.encode("utf-8")

            tpl = ((self.ChimieNelements, np.int32),
                   (SolarMassAbundances, np.float32), (labels, len(labels)))

            io.WriteBlock(f, tpl, byteorder=self.byteorder)

        # positions
        io.WriteArray(
            f,
            self.pos.astype(
                np.float32),
            byteorder=self.byteorder,
            pio=self.pio,
            npart=npart_write)
        # velocities
        io.WriteArray(
            f,
            self.vel.astype(
                np.float32),
            byteorder=self.byteorder,
            pio=self.pio,
            npart=npart_write)
        # id
        io.WriteArray(
            f,
            self.num.astype(
                np.int32),
            byteorder=self.byteorder,
            pio=self.pio,
            npart=npart_write)

        # write mass if needed
        if nzero != 0:
            io.WriteArray(
                f,
                massnzero.astype(
                    np.float32),
                byteorder=self.byteorder,
                pio=self.pio,
                npart=npart_m_write)

        # write extension
        if self.has_array('u'):
            if self.u is not None and self.npart[0] > 0:
                io.WriteArray(f, self.u[:self.npart[0]].astype(
                    np.float32), byteorder=self.byteorder, pio=self.pio, npart=None)
                self.message("write u")
        if self.has_array('rho'):
            if self.rho is not None and self.npart[0] > 0:
                io.WriteArray(f, self.rho[:self.npart[0]].astype(
                    np.float32), byteorder=self.byteorder, pio=self.pio, npart=None)
                self.message("write rho")
        if self.has_array('rsp'):
            if self.rsp is not None and self.npart[0] > 0:
                io.WriteArray(f, self.rsp[:self.npart[0]].astype(
                    np.float32), byteorder=self.byteorder, pio=self.pio, npart=None)
                self.message("write rsp")

        # this is the end of the minimal output

        if self.flag_metals:
            if self.has_array('metals'):
                io.WriteArray(f, self.metals[:self.npart[0]].astype(
                    np.float32), byteorder=self.byteorder, pio=self.pio, npart=None)
                self.message("write metals")

        if flag_thermaltime:
            if self.has_array('thtsnii'):
                io.WriteArray(f, self.thtsnii[:self.npart[0]].astype(
                    np.float32), byteorder=self.byteorder, pio=self.pio, npart=None)
                self.message("write thtsnii")

            if self.has_array('thtsnia'):
                io.WriteArray(f, self.thtsnia[:self.npart[0]].astype(
                    np.float32), byteorder=self.byteorder, pio=self.pio, npart=None)
                self.message("write thtsnia")

        if self.flag_age:
            if self.has_array('tstar') and self.npart[1] > 0:
                io.WriteArray(f, self.tstar[self.npart[0]:self.npart[0] + self.npart[1]].astype(
                    np.float32), byteorder=self.byteorder, pio=self.pio, npart=None)
                self.message("write tstar")

            if self.has_array('minit') and self.npart[1] > 0:
                io.WriteArray(f, self.minit[self.npart[0]:self.npart[0] + self.npart[1]].astype(
                    np.float32), byteorder=self.byteorder, pio=self.pio, npart=None)
                self.message("write minit")

            if self.has_array('idp') and self.npart[1] > 0:
                io.WriteArray(f, self.idp[self.npart[0]:self.npart[0] + self.npart[1]].astype(
                    np.int32), byteorder=self.byteorder, pio=self.pio, npart=None)
                self.message("write idp")

            if self.has_array('rho') and self.npart[1] > 0:
                data = self.rho[self.npart[0]:self.npart[0] +
                                self.npart[1]].astype(np.float32)
                if len(data) > 0:
                    io.WriteArray(
                        f, data, byteorder=self.byteorder,
                        pio=self.pio, npart=None)
                    self.message("write rho (stars)")

            if self.has_array('rsp') and self.npart[1] > 0:
                data = self.rsp[self.npart[0]:self.npart[0] +
                                self.npart[1]].astype(np.float32)
                if len(data) > 0:
                    io.WriteArray(
                        f, data, byteorder=self.byteorder,
                        pio=self.pio, npart=None)
                    self.message("write rsp (stars)")

        if self.flag_metals:
            if self.has_array('metals') and self.npart[1] > 0:
                data = self.metals[self.npart[0]:self.npart[0] +
                                   self.npart[1]].astype(np.float32)
                if len(data) > 0:
                    io.WriteArray(
                        f, data, byteorder=self.byteorder,
                        pio=self.pio, npart=None)
                    self.message("write metals (stars)")

        if self.has_array('pot'):
            if self.pot is not None:
                io.WriteArray(
                    f, self.pot.astype(np.float32),
                    byteorder=self.byteorder, pio=self.pio,
                    npart=None)
                self.message("write pot")

        if self.has_array('acc'):
            if self.acc is not None:
                io.WriteArray(
                    f, self.acc.astype(np.float32),
                    byteorder=self.byteorder,
                    pio=self.pio, npart=None)
                self.message("write acc")

        if self.has_array('opt1'):
            if self.opt1 is not None:
                io.WriteArray(f, self.opt1[:self.npart[0]].astype(
                    np.float32), byteorder=self.byteorder, pio=self.pio, npart=None)
                self.message("write opt1")

        if self.has_array('opt2'):
            if self.opt2 is not None:
                io.WriteArray(f, self.opt2[:self.npart[0]].astype(
                    np.float32), byteorder=self.byteorder, pio=self.pio, npart=None)
                self.message("write opt2")

        if self.has_array('erd'):
            if self.erd is not None:
                io.WriteArray(f, self.erd[:self.npart[0]].astype(
                    np.float32), byteorder=self.byteorder, pio=self.pio, npart=None)
        if self.has_array('dte'):
            if self.dte is not None:
                io.WriteArray(f, self.dte[:self.npart[0]].astype(
                    np.float32), byteorder=self.byteorder, pio=self.pio, npart=None)
